# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: odedynamics.py,v 1.8 2006/05/26 17:08:04 mbaas Exp $

## \file odedynamics.py
## Contains the ODEDynamics class.

"""This module contains a Dynamics component using the ODE rigid body
dynamics package."""

import weakref
from . import protocols
from .Interfaces import *
from .component import Component
from .globalscene import getScene
from .eventmanager import eventManager
from .cgtypes import *
from .worldobject import WorldObject
from .spheregeom import SphereGeom
from .ccylindergeom import CCylinderGeom
from .boxgeom import BoxGeom
from .trimeshgeom import TriMeshGeom
from .planegeom import PlaneGeom
from .joints import *
from .events import *
from .slots import *
from . import cmds
from . import _core
try:
    import ode
    has_ode = True
except:
    has_ode = False

#import numarray
#import numarray.linear_algebra

ODE_COLLISION = "ODECollision"

# ODECollisionEvent
class ODECollisionEvent:
    """Event object that is used as argument for collision events.
    """
    
    def __init__(self, obj1, obj2, contacts, contactproperties):
        self.obj1 = obj1
        self.obj2 = obj2
        self.contacts = contacts
        self.contactproperties = contactproperties

    def __str__(self):
        if self.obj1==None:
            n1 = "None"
        else:
            n1 = '"%s"'%self.obj1.name
        if self.obj2==None:
            n2 = "None"
        else:
            n2 = '"%s"'%self.obj2.name
        return '<ODECollisionEvent: obj1:%s obj2:%s #Contacts:%d>'%(n1, n2, len(self.contacts))

    # averageContactGeom
    def averageContactGeom(self):
        """Return the average contact position, normal and depth.
        """
        pos = vec3(0)
        normal = vec3(0)
        depth = 0.0
        for c in self.contacts:
            p,n,d,g1,g2 = c.getContactGeomParams()
            pos += vec3(p)
            normal += vec3(n)
            depth += d

        # n is not 0 (the event wouldn't have been generated if it was)
        n = len(self.contacts)
        pos /= n
        normal /= n
        depth /= n

        return pos,normal,depth
    

# ODEDynamics
class ODEDynamics(Component):
    """Dynamics component using the ODE rigid body dynamics package.
    """

    def __init__(self,
                 name="ODEDynamics",
                 gravity = 9.81,
                 substeps = 1,
                 cfm = None,
                 erp = None,
                 defaultcontactproperties = None,
                 enabled = True,
                 show_contacts = False,
                 contactmarkersize = 0.1,
                 contactnormalsize = 1.0,
                 collision_events = False,
                 use_quick_step = True,
                 auto_add = False,
                 auto_insert = True):
        """Constructor.

        \param name (\c str) Component name
        \param use_quick_step (\c bool) if True, use QuickStep method for ODE
               stepping (default); if False, use Step (slower, but more accurate)
        \param auto_add (\c bool) Automatically add the world objects to the simulation
        \param auto_insert (\c bool) Automatically add the component to the scene
        """
        Component.__init__(self, name=name, auto_insert=auto_insert)

        scene = getScene()

        self.use_quick_step = use_quick_step
        
        self.substeps = substeps
        self.collision_events = collision_events

        self.world = ode.World()
        g = -gravity*scene.up
        self.world.setGravity(g)
        if cfm!=None:
            self.world.setCFM(cfm)
        if erp!=None:
            self.world.setERP(erp)
        
#        self.world.setAutoDisableFlag(True)

        self.enabled = enabled

        self.eventmanager = eventManager()
        self.eventmanager.connect(STEP_FRAME, self)
        self.eventmanager.connect(RESET, self.reset)

        # Space object
        self.space = ode.Space()

        # Joint group for the contact joints
        self.contactgroup = ode.JointGroup()

        # A dictionary with WorldObjects as keys and ODEBodies as values
        self.body_dict = {}
        # A list of all bodies
        self.bodies = []
        # A list of all joints
        self.joints = []

        # A dictionary with contact properties
        # Key is a 2-tuple (material1, material2). Value is a
        # ODEContactProperties object.
        # The value of (mat1, mat2) should always be the same than
        # the value of (mat2, mat1).
        self.contactprops = {}

        # Default contact properties
        if defaultcontactproperties==None:
            defaultcontactproperties = ODEContactProperties()
        self.defaultcontactprops = defaultcontactproperties

        self.show_contacts = show_contacts
        self.contactmarkersize = contactmarkersize
        self.contactnormalsize = contactnormalsize

        # A list of weakrefs to body manipulators
        self.body_manips = []

        # Debug statistics (the number of contacts per simulation step)
        self.numcontacts = 0

        # Automatically add world objects
        if auto_add:
            # Add all rigid bodies first...
            for obj in scene.worldRoot().iterChilds():
                try:
                    obj = protocols.adapt(obj, IRigidBody)
                    self.add(obj)
                except NotImplementedError:
                    pass

            # Then add all joints...
            for obj in scene.worldRoot().iterChilds():
                if isinstance(obj, ODEJointBase):
                    self.add(obj)

    # remove
    def remove(self, objs):
        """Remove one or more world objects from the simulation.

        \param objs  World objects to be removed (given as names or objects)
        """
        objs = cmds.worldObjects(objs)
        for obj in objs:
            self._remove(obj)
    
    def _remove(self, obj):
        for body in self.bodies:
            if body.obj == obj:
                # after deleting all the references to the body, 
                # PyODE will remove it from simulation (in the destructor)
                self.bodies.remove(body)
                del(self.body_dict[obj])
                del(body.odegeoms)
                del(body.odebody)
                del(obj.manip)
                break
        

    # add
    def add(self, objs, categorybits=None, collidebits=None):
        """Add a world object to the simulation.

        \param objs World objects to be simulated (given as names or objects)
        """
        objs = cmds.worldObjects(objs)
        for obj in objs:
            self._add(obj, categorybits, collidebits)

    def _add(self, obj, categorybits, collidebits):
        
        if isinstance(obj, ODEJointBase):
            obj.activate(self)
            self.joints.append(obj)
            return

        try:
            obj = protocols.adapt(obj, IRigidBody)
        except NotImplementedError:
            print(('Object "%s" is not a rigid body.'%obj.name))
            return

        # Should the object be ignored?
        if not obj.dynamics:
            return

#       print "#####################",obj.name,"#########################"
#       print "Old center of gravity:",obj.cog
#       print "Old inertiatensor:"
#        print obj.inertiatensor
#       print "Old Offset transform:"
        P = obj.getOffsetTransform()
#       print P
        obj.pivot = P*obj.cog
#       print "Setting pivot to", obj.pivot
#        print "New center of gravity:", obj.cog
#        print "Intermediate inertia tensor:"
#        print obj.inertiatensor
#        I = obj.inertiatensor
#        a = numarray.array([I.getRow(0), I.getRow(1), I.getRow(2)], type=numarray.Float64)
#        vals, vecs = numarray.linear_algebra.eigenvectors(a)
#        print "Eigenvalues:",vals
        # Eigenvektoren sind ungenau! (bei cube_base_cog grosse Abweichungen)
#        b1 = vec3(vecs[0]).normalize()
#        b2 = vec3(vecs[1]).normalize()
#        b3 = vec3(vecs[2]).normalize()
#        P = obj.getOffsetTransform()        
#        Pnull = P*vec3(0,0,0)
#        b1 = P*b1 - Pnull
#        b2 = P*b2 - Pnull
#        b3 = P*b3 - Pnull        
#        I2 = mat3(b1,b2,b3)
#        print I2
#        if I2.determinant()<0:
#            I2.setColumn(0, -I2.getColumn(0))
#        print "Det of new basis",I2.determinant()
#        P = obj.getOffsetTransform()
#        P.setMat3(I2)
#        obj.setOffsetTransform(P)
#        print "New offset transform:"
#        print P
#        print "New center of gravity:",obj.cog
#        print "New inertia tensor:"
#        print obj.inertiatensor


        # If the mass of a rigid body is <= 0, ODE will crash
        # Using a more reasonable default: mass = 1
        if obj.mass <= 0:
            print(("Using default mass=1 for " + str(obj)))
            obj.mass = 1
        body = ODEBody(obj, self, categorybits=categorybits, collidebits=collidebits)
        self.bodies.append(body)
        self.body_dict[obj] = body

        obj.manip = self.createBodyManipulator(obj) # Quick access to manipulator

    # reset
    def reset(self):
        """Reset the body states.

        All bodies are reset to their position and velocity at the time
        they were added.
        
        This method is also called when the RESET event is issued.
        """
        for b in self.bodies:
            b.reset()
            b.updateObj()

        for j in self.joints:
            j.reset()

    # setContactProperties
    def setContactProperties(self, mat_tuple, props):
        """Set the contact properties of a pair of materials.

        The contact properties \a props are applied whenever an object
        with material \a mat1 collides with an object with material \a mat2.

        \param mat1 (\c Material) Material 1
        \param mat2 (\c Material) Material 2
        \param props (\c ODEContactProperties) Contact properties
        """
        mat1,mat2 = mat_tuple
        self.contactprops[(mat1,mat2)] = props  
        # Collision events are forced to appear in (mat1,mat2) order
     
    # getContactProperties
    def getContactProperties(self, matpair):
        """Return the contact properties for a material pair.

        \param matpair A 2-tuple of two Material objects
        \return Contact properties (\c ODEContactProperties)
        """
        cp = self.contactprops.get(matpair)
        if cp==None:
#            print 'ODEDynamics: Warning: No contact properties defined for material "%s" and "%s"'%(matpair[0].name, matpair[1].name)
            cp = self.defaultcontactprops
        return cp

    # createBodyManipulator
    def createBodyManipulator(self, obj):
#        return self.body_dict[obj].odebody
        bm = ODEBodyManipulator(self.body_dict[obj])
        self.body_manips.append(weakref.ref(bm))
        return bm

    # nearCallback
    def nearCallback(self, args, geom1, geom2):        
        try:            
            # Force collision event to appear in (mat1, mat2) order
            # (i.e. the order in which the material pair was defined in contactprops)
            if not (geom1.material, geom2.material) in self.contactprops:
                (geom1, geom2) = (geom2, geom1)
        except:
            pass

        # Check if the objects do collide
        contacts = ode.collide(geom1, geom2)

        # No contacts? then return immediately
        if len(contacts)==0:
            return
        
#        print len(contacts),"contacts"
        self.numcontacts += len(contacts)

        # Get the contact properties
        cp = self.getContactProperties((geom1.material, geom2.material))

        # Create a collision event?
        if self.collision_events:
            obj1 = getattr(geom1, "worldobj", None)
            obj2 = getattr(geom2, "worldobj", None)
            evt = ODECollisionEvent(obj1, obj2, contacts, cp)
            self.eventmanager.event(ODE_COLLISION, evt)

        # Create contact joints
        for c in contacts:
            if self.show_contacts:
                p,n,d,g1,g2 = c.getContactGeomParams()
                cmds.drawMarker(p, size=self.contactmarkersize, color=(1,0,0))
                cmds.drawLine(p, vec3(p)+self.contactnormalsize*vec3(n), color=(1,0.5,0.5))
#                print p,n,d
            
            # Set the properties
            cp.apply(c)
            # Create the contact joint
            j = ode.ContactJoint(self.world, self.contactgroup, c)
            b1 = geom1.getBody()
            b2 = geom2.getBody()
#            if b1==None:
#                b1=ode.environment
#            if b2==None:
#                b2=ode.environment
            j.attach(b1, b2)

    # onStepFrame
    def onStepFrame(self):
        """Callback for the StepFrame event.
        """
        if self.substeps==0 or not self.enabled:
            return

        if self.show_contacts:
            cmds.drawClear()

        # Remove dead body manipulators...
        self.body_manips = [x for x in self.body_manips if x() is not None]

        # Sim loop...
        subdt = getScene().timer().timestep/self.substeps
        for i in range(self.substeps):
            self.numcontacts = 0
            
            # Apply body manipulators
            for bmref in self.body_manips:
                bm = bmref()
                if bm is not None:
                    bm._apply()
            
            # Detect collisions and create contact joints
            self.space.collide(None, self.nearCallback)
#            print "#Contacts:",self.numcontacts

            # Simulation step
            if self.use_quick_step:
                self.world.quickStep(subdt)
            else:
                self.world.step(subdt)

            # Remove all contact joints
            self.contactgroup.empty()
            
        # Update the world objects
        for body in self.bodies:
            body.updateObj()

        # Reset body manipulators
        for bmref in self.body_manips:
            bm = bmref()
            if bm is not None:
                bm._reset()


######################################################################

# ODEContactProperties
class ODEContactProperties:

    """Contact properties.

    This class stores contact properties that are used whenever
    two objects collide. The attributes are used to initialize
    ODE contact joints.
    """    
    
    def __init__(self,
                 mode = 0,
                 mu = 0.3,
                 mu2 = None,
                 bounce = None,
                 bounce_vel = None,
                 soft_erp = None,
                 soft_cfm = None,
                 motion1 = None,
                 motion2 = None,
                 slip1 = None,
                 slip2 = None,
                 fdir1 = None):
        """Constructor.

        See the ODE manual for an explanation of the parameters.

        The flags for the mode parameter are automatically set.
        However, you can initially set the ContactApprox flags.
        """

        if mu2!=None:
            mode |= ode.ContactMu2
        else:
            mu2 = 0.0

        if bounce!=None:
            mode |= ode.ContactBounce
        else:
            bounce = 0.0

        if bounce_vel==None:
            bounce_vel = 0.0

        if soft_erp!=None:
            mode |= ode.ContactSoftERP
        else:
            soft_erp = 0.0

        if soft_cfm!=None:
            mode |= ode.ContactSoftCFM
        else:
            soft_cfm = 0.0

        if motion1!=None:
            mode |= ode.ContactMotion1
        else:
            motion1 = 0.0

        if motion2!=None:
            mode |= ode.ContactMotion2
        else:
            motion2 = 0.0

        if slip1!=None:
            mode |= ode.ContactSlip1
        else:
            slip1 = 0.0

        if slip2!=None:
            mode |= ode.ContactSlip2
        else:
            slip2 = 0.0

        if fdir1!=None:
            mode |= ode.ContactFDir1
        else:
            fdir1 = (0,0,0)

        self.mode = mode
#        print "ODEContactProps: mode =",mode
        self.mu = mu
        self.mu2 = mu2
        self.bounce = bounce
        self.bounce_vel = bounce_vel
        self.soft_erp = soft_erp
        self.soft_cfm = soft_cfm
        self.motion1 = motion1
        self.motion2 = motion2
        self.slip1 = slip1
        self.slip2 = slip2
        self.fdir1 = fdir1

    # apply
    def apply(self, contact):
        """Apply the contact properties to a contact joint."""
        contact.setMode(self.mode)
        contact.setMu(self.mu)
        contact.setMu2(self.mu2)
        contact.setBounce(self.bounce)
        contact.setBounceVel(self.bounce_vel)
        contact.setMotion1(self.motion1)
        contact.setMotion2(self.motion2)
        contact.setSlip1(self.slip1)
        contact.setSlip2(self.slip2)
        contact.setSoftERP(self.soft_erp)
        contact.setSoftCFM(self.soft_cfm)
        contact.setFDir1(self.fdir1)
                         

######################################################################

# ODEBodyManipulator
class ODEBodyManipulator(object):
    """Body manipulator class.

    This class can be used to apply external forces/torques to a body
    or to manipulate the position/orientation/velocities directly.

    You should not create instances of this class yourself. Instead, use
    the createBodyManipulator() method of the ODEDynamics component.
    """
    
    def __init__(self, body):
        """Constructor.

        \param body (\c ODEBody) The internal body object representing the
               rigid body to manipulate.
        """
        self._body = body
        self._odebody = body.odebody
        self._reset()

    # setPos
    def setPos(self, pos):
        """Set the position of the body.

        pos must be a 3-sequence of floats.
        """
        self._odebody.enable()
        self._odebody.setPosition(pos)

    # setRot
    def setRot(self, rot):
        """Set the rotation of the body.

        rot must be a mat3 containing the orientation or a list of 9 values
        in row-major order.
        """
        self._odebody.enable()
        self._odebody.setRotation(mat3(rot).toList(True))  # Now setRot really accepts a mat3 (and also a list with 9 elements)

    # setLinearVel
    def setLinearVel(self, vel):
        """Set the linear velocity.

        vel must be a 3-sequence of floats.
        """
        self._odebody.enable()
        self._odebody.setLinearVel(vel)

    # setAngularVel
    def setAngularVel(self, vel):
        """Set the angular velocity.

        vel must be a 3-sequence of floats.
        """
        self._odebody.enable()
        self._odebody.setAngularVel(vel)

    # setInitialPos
    def setInitialPos(self, pos):
        """Set the initial position.

        pos must be a 3-sequence of floats.
        """
        self._body.initial_pos = vec3(pos)

    # setInitialRot
    def setInitialRot(self, rot):
        """Set the initial orientation.

        rot must be a mat3.
        """
        self._body.initial_rot = mat3(rot)

    # setInitialLinearVel
    def setInitialLinearVel(self, vel):
        """Set the initial linear velocity.

        vel must be a 3-sequence of floats.
        """
        self._odebody.initial_linearvel = vec3(vel)

    # setInitialAngularVel
    def setInitialAngularVel(self, vel):
        """Set the initial angular velocity.

        vel must be a 3-sequence of floats.
        """
        self._odebody.initial_angularvel = vec3(vel)

    # addForce
    def addForce(self, force, relforce=False, pos=None, relpos=False):
        """Apply an external force to a rigid body.
        
        Add an external force to the current force vector. force is a
        vector containing the force to apply. If relforce is True the
        force is interpreted in local object space, otherwise it is
        assumed to be given in global world space. By default, the
        force is applied at the center of gravity. You can also pass a
        different position in the pos argument which must describe a
        point in space. relpos determines if the point is given in
        object space or world space (default).
        """
        
        R = None
        force = vec3(force)
        if relforce:
            R = mat3(self._odebody.getRotation())
            force = R*force
        # Is a position given? then add the corresponding torque
        if pos!=None:
            pos = vec3(pos)
            bodyorig = vec3(self._odebody.getPosition())
            if relpos:
                if R==None:
                    R = mat3(self._odebody.getRotation())
                pos = R*pos + bodyorig

            self._torque += (pos-bodyorig).cross(force)
            self._torque_flag = True
            
        self._force += vec3(force)
        self._force_flag = True
                    
    # addTorque
    def addTorque(self, torque, reltorque=False):
        """Apply an external torque to a rigid body.
        
        Add an external torque to the current torque vector. torque is
        a vector containing the torque to apply. reltorque determines
        if the torque vector is given in object space or world space
        (default).
        """
        
        torque = vec3(torque)
        if reltorque:
            R = mat3(self._odebody.getRotation())
            torque = R*torque
            
        self._torque += torque
        self._torque_flag = True

    # _apply
    def _apply(self):
        """Apply the stored force/torque.

        This method is called by the ODEDynamics object during the simulation
        step (once for every sub step).
        """
        
        if self._force_flag or self._torque_flag:
            self._odebody.enable()
            
        if self._force_flag:
            self._odebody.addForce(self._force)
        if self._torque_flag:
            self._odebody.addTorque(self._torque)

    # _reset
    def _reset(self):
        """Reset the stored force/torque.

        This method is called by the ODEDynamics object at the end of one
        entire simulation step.
        """
        self._force = vec3(0)
        self._torque = vec3(0)
        
        self._force_flag = False
        self._torque_flag = False

    ## protected:
        
    # "body" property...
    
    def _getBody(self):
        """Return the current body (WorldObject).

        This method is used for retrieving the \a body property.

        \return Rigid body
        """
        return self._body.obj

    body = property(_getBody, None, None, "Rigid body")

    # "odebody" property...
    # Someone may think this method would return an ODEBody, not an ode.Body
    # That's why I changed the case.
    def _get_odeBody(self):
        """Return the current ODE body (type ode.Body from PyODE). 

        This method is used for retrieving the \a odebody property.

        \return ODE body
        """
        return self._odebody

    odebody = property(_get_odeBody, None, None, "ODE body")

    # odegeoms property
    
    def _get_odeGeoms(self):
        """Return the current ODE geom list.

        This method is used for retrieving the \a odegeoms property.

        \return ODE geom list
        """
        return self._body.odegeoms

    odegeoms = property(_get_odeGeoms, None, None, "ODE geoms")

    
    
######################################################################

# ODEBody
class ODEBody:
    """This class is the interface between ODE bodies and world objects.

    It encapsulates the ODE body and the ODE geom objects.

    """
    def __init__(self, obj, dyncomp, categorybits=None, collidebits=None):
        """Constructor.

        \param obj (\c WorldObject) World object
        \param dyncomp (\c ODEDynamics) Dynamics component
        """
        
        # Store the world object
        self.obj = obj

        self.dyncomp = dyncomp
        world = self.dyncomp.world
        space = self.dyncomp.space

        # Store initial position/orientation
        self.initial_pos = obj.pos
        self.initial_rot = obj.rot
        self.initial_linearvel = obj.linearvel #vec3(0,0,0)
        self.initial_angularvel = obj.angularvel #vec3(0,0,0)

        # The ODE body
        self.odebody = None
        # A list of ODE geoms 
        self.odegeoms = None
    
        # Create ODE collision geoms
        self.odegeoms = self.createGeoms(obj, space)

        # Set the category and collide bits
        for geom in self.odegeoms:
            if categorybits!=None:
                geom.setCategoryBits(categorybits)
            if collidebits!=None:
                geom.setCollideBits(collidebits)

        # Create an ODE body
        if not obj.static:
            self.odebody = self.createBody(obj, world)
            for g in self.odegeoms:
                g.setBody(self.odebody)

        # Store the material in the ODE geoms so that the contact
        # properties can be determined in the near callback.
        # Also store the corresponding WorldObject
        for g in self.odegeoms:
            g.material = obj.getMaterial()
            g.worldobj = obj

        # Initialize the ODE body/geoms
        self.reset()

        # Create the notification forwarders and make connections
        self.initialization = True
        # Forward changes to the "static" slot
        self._static_forwarder = NotificationForwarder(self.onStaticChanged)
        self.obj.static_slot.addDependent(self._static_forwarder)
        # Forward changes to the "mass" slot
        self._mass_forwarder = NotificationForwarder(self.onMassChanged)
        self.obj.mass_slot.addDependent(self._mass_forwarder)
        self.initialization = False

    # updateObj
    def updateObj(self):
        """Update the transformation of the world object.

        The transformation from the ODE body is copied to the corresponding
        world object. The linearvel and angularvel attributes are
        updated as well.
        """
        if self.obj.static:
            return

#        if type(self.odegeom)==ode.GeomTriMesh:
#            self.odegeom.clearTCCache()

        # Get the ODE transformation
        pos = self.odebody.getPosition()
        rot = self.odebody.getRotation()
        # Create the transformation matrix for the world object
        M = mat4().translation(vec3(pos))
        M.setMat3(mat3(rot))
        # Set the transformation on the world object
        self.obj.transform = M

        self.obj.linearvel = vec3(self.odebody.getLinearVel())
        self.obj.angularvel = vec3(self.odebody.getAngularVel())

    ## protected:

    # onStaticChanged
    def onStaticChanged(self):
        if self.initialization:
            return
        
        if self.obj.static:
            if self.odebody!=None:
                for g in self.odegeoms:
                    g.setBody(None)
            self.odebody = None
        else:
            self.odebody = self.createBody(self.obj, self.dyncomp.world)
            for g in self.odegeoms:
                g.setBody(self.odebody)
            self.odebody.setPosition(self.obj.pos)
            self.odebody.setRotation(self.obj.rot.toList(rowmajor=True))
            self.odebody.setLinearVel((0,0,0))
            self.odebody.setAngularVel((0,0,0))            
            
    # onMassChanged
    def onMassChanged(self):
        if self.initialization:
            return
        
        print(("Mass changed to",self.obj.mass))

    # reset
    def reset(self):
        """Restore the initial state of the body."""
        if self.obj.static:
            for g in self.odegeoms:
                if g.placeable():
                    g.setPosition(self.initial_pos)
                    g.setRotation(self.initial_rot.toList(rowmajor=True))
        else:
            self.odebody.setPosition(self.initial_pos)
            self.odebody.setRotation(self.initial_rot.toList(rowmajor=True))
            self.odebody.setLinearVel(self.initial_linearvel)
            self.odebody.setAngularVel(self.initial_angularvel)

    # createBody
    def createBody(self, obj, world):
        """Create an ODE body.

        Creates an ODE body and initializes the mass properties.
        """
        # Create an ODE Mass object
        M = ode.Mass()
        m = obj.totalmass
        if m==0:
            print(('WARNING: Object "%s" has a mass of zero.'%obj.name))
        cog = obj.cog
        I = obj.inertiatensor
#        print '---Rigid body "%s"--------'%obj.name
#        print 'cog:',cog
#        print I
        I = mat3(I)
        M.setParameters(m, cog.x, cog.y, cog.z, I[0,0], I[1,1], I[2,2], I[0,1], I[0,2], I[1,2])

        # Create an ODE body
        body = ode.Body(world)
        body.setMass(M)

        return body

    # createGeoms
    def createGeoms(self, obj, space):
        """Create all the geoms for one world object including the children.

        The generated geoms will be encapsulated in GeomTransforms
        so that all of them are defined with respect to the pivot coordinate
        system of \a obj.

        \param obj (\c WorldObject) Top level world object
        \param space (\c ode.Space) ODE Space object
        \return List of ODE geom objects
        """

        # Plane? This is a special case because ODE planes are not placeable
        if isinstance(obj.geom, PlaneGeom):
            if not obj.static:
                raise RuntimeError("Planes can only be used as static bodies")
            L = obj.localTransform()
            n = L*vec3(0,0,1) - L*vec3(0)
            n = n.normalize()
            d = n*obj.pos
            odegeom = ode.GeomPlane(space, n, d)
            return [odegeom]

        res = []

        # Create all ODE geoms in the subtree starting at obj
        # Each geom will have a position/rotation that moves it to
        # the local coordinate system of obj.
        geoms = self._createLGeoms(obj, mat4(1))

        # Apply the offset transform and encapsulate the geoms inside
        # geom transforms...
        P = obj.getOffsetTransform()
        Pinv = P.inverse()
        pos,rot,scale = Pinv.decompose()
        if scale!=vec3(1,1,1):
            print('WARNING: ODEDynamics: Scaled geometries are not supported')
        res = []
        for g in geoms:
            M = mat4(1).translation(vec3(g.getPosition()))
            M.setMat3(mat3(g.getRotation()))
            M *= Pinv;
            p4 = M.getColumn(3)
            g.setPosition((p4.x, p4.y, p4.z))
            # row major or column major?
            g.setRotation(M.getMat3().toList(rowmajor=True))
        
            gt = ode.GeomTransform(space)
            gt.setGeom(g)
            res.append(gt)

        return res
        
    # _createLGeoms
    def _createLGeoms(self, obj, M):
        """Create ODE geom objects for a world object.

        Create the ODE geoms for \a obj (and its children).
        The transformation \a M is applied to each geom.

        The returned geoms are not assigned to a Body yet.

        \param obj (\c WorldObject) The world object
        \param M (\c mat4) Transformation
        \return List of ODE geoms
        """
        res = []
        # Create an ODE geom for the object itself
        geom = obj.geom
        if geom!=None:
            res.append(self._createGeom(geom, M))

        # Create ODE geoms for the children
        for child in obj.iterChilds():
            if not getattr(child, "dynamics", False):
                continue
            L = child.localTransform()
            res += self._createLGeoms(child, M*L)

        return res
        
    # _createGeom
    def _createGeom(self, geom, M):
        """Create an ODE geom object for the given cgkit geom.

        Create an ODE geom of the appropriate type (depening on \a geom)
        and apply the transformation \a M to it (currently, M may only
        contain a translation and a rotation, otherwise a warning is
        printed).

        The returned geom is not assigned to a Body yet.

        \param geom (\c GeomObject) cgkit geom object
        \param M (\c mat4) Transformation that's applied to the geom
        \return ODE geom
        """

        # Create the raw geom object that's defined in the local
        # coordinate system L of the world object this geom belongs to.

        # Sphere geom?
        if isinstance(geom, SphereGeom):
            odegeom = ode.GeomSphere(None, geom.radius)
        # CCylinder geom?
        elif isinstance(geom, CCylinderGeom):
            odegeom = ode.GeomCCylinder(None, geom.radius, geom.length)
        # Box geom?
        elif isinstance(geom, BoxGeom):
            odegeom = ode.GeomBox(None, (geom.lx, geom.ly, geom.lz))
        # TriMesh geom?
        elif isinstance(geom, TriMeshGeom):
            verts = []
            for i in range(geom.verts.size()):
                verts.append(geom.verts.getValue(i))
#                print "V",i,verts[i]
            faces = []
            for i in range(geom.faces.size()):
                f = geom.faces.getValue(i)
                faces.append(f)
#                print "F",i,faces[i]
                ia,ib,ic = faces[i]
                a = verts[ia]
                b = verts[ib]
                c = verts[ic]
                sa = ((b-a).cross(c-a)).length()
#                if sa<0.0001:
#                    print "*****KLEIN*****",sa

            tmd = ode.TriMeshData()
            tmd.build(verts, faces)            
            odegeom = ode.GeomTriMesh(tmd, None)
        # Plane geom?
        elif isinstance(geom, PlaneGeom):
            L = obj.localTransform()
            n = L*vec3(0,0,1) - L*vec3(0)
            n = n.normalize()
            d = n*obj.pos
            odegeom = ode.GeomPlane(space, n, d)
        # Unknown geometry
        else:
            raise ValueError('WARNING: ODEDynamics: Cannot determine collision geometry of object "%s".'%geom.name)
#            print 'WARNING: ODEDynamics: Cannot determine collision geometry of object "%s". Using bounding box instead.'%obj.name
#            bmin, bmax = obj.boundingBox().getBounds()
#            s = bmax-bmin
#            odegeom = ode.GeomBox(None, s)
#            pos,rot,scale = P.inverse().decompose()
#            odegeom.setPosition(pos + 0.5*(bmax+bmin))
#            odegeomtransform = ode.GeomTransform(space)
#            odegeomtransform.setGeom(odegeom)
#            return odegeomtransform

        # Displace the geom by M
        pos,rot,scale = M.decompose()
        if scale!=vec3(1,1,1):
            print('WARNING: ODEDynamics: Scaled geometries are not supported')
        odegeom.setPosition(pos)
        # row major or column major?
        odegeom.setRotation(rot.getMat3().toList(rowmajor=True))
        
        return odegeom

######################################################################

# ODEJointBase
class ODEJointBase(WorldObject):

    protocols.advise(instancesProvide=[ISceneItem])

    exec(slotPropertyCode("lostop"))
    exec(slotPropertyCode("histop"))
    exec(slotPropertyCode("motorvel"))
    exec(slotPropertyCode("motorfmax"))
    exec(slotPropertyCode("fudgefactor"))
    exec(slotPropertyCode("bounce"))
    exec(slotPropertyCode("cfm"))
    exec(slotPropertyCode("stoperp"))
    exec(slotPropertyCode("stopcfm"))

    def __init__(self,
                 name = "",
                 body1 = None,
                 body2 = None,
                 **params):
        WorldObject.__init__(self, name=name, **params)
        
        # ODEDynamics component
        self.odedynamics = None

        # The corresponding ODE joint
        self.odejoint = None

        self.body1 = body1
        self.body2 = body2

    # attach
    def attach(self, body1, body2=None):
        self.body1 = body1
        self.body2 = body2
        if self.odejoint!=None:
            if body1==None:
                b1 = ode.environment
            else:
                b1 = self.odedynamics.body_dict[body1].odebody
            
            if body2==None:
                b2 = ode.environment
            else:
                b2 = self.odedynamics.body_dict[body2].odebody

            self.odejoint.attach(b1, b2)

    # activate
    def activate(self, odedynamics):
        """
        this method is called by the ODEDynamics componenent
        """
        if odedynamics==self.odedynamics:
            return
        if self.odedynamics!=None:
            print(('Warning: Joint "%s" is already in use'%self.name))
        self.odedynamics = odedynamics
        self._createODEjoint()
        self._initODEjoint()

    # reset
    def reset(self):
        """Reset method.

        This may only be called after the bodies have been reset.
        """
        pass

    def _createSlots(self):
        
        self._createSlot("lostop", None, "-ode.Infinity", "onLoStopChanged")
        self._createSlot("histop", None, "ode.Infinity", "onHiStopChanged")
        self._createSlot("motorvel", None, "0.0", "onMotorVelChanged")
        self._createSlot("motorfmax", None, "0.0", "onMotorFMaxChanged")
        self._createSlot("fudgefactor", None, "1.0", "onFudgeFactorChanged")
        self._createSlot("bounce", None, "0.1", "onBounceChanged")
        self._createSlot("cfm", None, 1E-5, "onCFMChanged")
        self._createSlot("stoperp", None, 0.2, "onStopERPChanged")
        self._createSlot("stopcfm", None, 1E-5, "onStopCFMChanged")
        
    def _createSlot(self, name, nr, default, callback):
        """Create one joint param slot.

        This method creates a DoubleSlot called \a name (with additional
        number, if \a nr is not None). \a default is the initial default
        value (as string) and \a callback the name of the callback method.

        \param name (\c str) Parameter name
        \param nr (\c int) Parameter set number or None
        \param default (\c str) Default value as string
        \param callback (\c str) Name of the callback method
        """
        if nr!=None:
            name += str(nr)
        # Create the slot
        s = "self.%s_slot = DoubleSlot(%s)"%(name, default)
        exec(s)
        # Add the slot to the component
        s = "self.addSlot('%s', self.%s_slot)"%(name, name)
        exec(s)
        # Create the forwarder
        s = "self._%s_forwarder = NotificationForwarder(self.%s)"%(name, callback)
        exec(s)
        s = "self.%s_slot.addDependent(self._%s_forwarder)"%(name, name)
        exec(s)
       

    # _createODEjoint
    def _createODEjoint(self):
        """Create the corresponding ODE joint.

        This may only be called if the ODEDynamics component has been set.

        This method has to create an ODE joint object and store it
        in self.odejoint.
        """
        pass

    # _initODEjoint
    def _initODEjoint(self):
        """Initialize the ODE joint.
        """
        # This will call the attach() method of the ODE joint
        self.attach(self.body1, self.body2)
        
        try:
            print(("***",self.odejoint.getParam(ode.ParamStopCFM)))
            
            self.odejoint.setParam(ode.ParamFMax, self.motorfmax)
            self.odejoint.setParam(ode.ParamVel, self.motorvel)
            self.odejoint.setParam(ode.ParamLoStop, self.lostop)
            self.odejoint.setParam(ode.ParamHiStop, self.histop)
            self.odejoint.setParam(ode.ParamFudgeFactor, self.fudgefactor)
            self.odejoint.setParam(ode.ParamBounce, self.bounce)
            self.odejoint.setParam(ode.ParamCFM, self.cfm)
            self.odejoint.setParam(ode.ParamStopERP, self.stoperp)
            self.odejoint.setParam(ode.ParamStopCFM, self.stopcfm)
        except AttributeError:
            pass  # not all joints have these attributes

    def onLoStopChanged(self):
#        print "Lostop has been changed to",self.lostop_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamLoStop, self.lostop_slot.getValue())

    def onHiStopChanged(self):
#        print "Histop has been changed to",self.histop_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamHiStop, self.histop_slot.getValue())

    def onMotorVelChanged(self):
#        print "Motor velocity has been changed to",self.motorvel_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamVel, self.motorvel)


    def onMotorFMaxChanged(self):
#        print "Motor FMax has been changed to",self.motorfmax_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamFMax, self.motorfmax)

    def onFudgeFactorChanged(self):
#        print "Fudgefactor has been changed to",self.fudgefactor_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamFudgeFactor, self.fudgefactor_slot.getValue())

    def onBounceChanged(self):
#        print "Bounce has been changed to",self.bounce_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamBounce, self.bounce_slot.getValue())


    def onMotorVel2Changed(self):
#        print "Motor2 velocity has been changed to",self.motorvel2_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamVel2, self.motorvel2)


    def onMotorFMax2Changed(self):
#        print "Motor2 FMax has been changed to",self.motorfmax2_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamFMax2, self.motorfmax2)

    def onCFMChanged(self):
#        print "CFM has been changed to",self.cfm_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamCFM, self.cfm)

    def onStopCFMChanged(self):
#        print "Stop CFM has been changed to",self.stopcfm_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamStopCFM, self.stopcfm)

    def onStopERPChanged(self):
#        print "Stop ERP has been changed to",self.stoperp_slot.getValue()
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamStopERP, self.stoperp)


# BallJoint
class ODEBallJoint(ODEJointBase):

    def __init__(self,
                 name = "ODEBallJoint",
                 body1 = None,
                 body2 = None,
                 **params):
        ODEJointBase.__init__(self,
                              name=name, body1=body1, body2=body2,
                              **params)

        self._createSlots()

    # _createODEjoint
    def _createODEjoint(self):
        # Create the ODE joint
        self.odejoint = ode.BallJoint(self.odedynamics.world)

    # _initODEjoint
    def _initODEjoint(self):
        ODEJointBase._initODEjoint(self)

        W = self.worldtransform
        p = W[3]
        self.odejoint.setAnchor((p.x, p.y, p.z))


# HingeJoint
class ODEHingeJoint(ODEJointBase):
    """
    the rotation axis is the local x axis
    """

    exec(slotPropertyCode("angle"))

    def __init__(self,
                 name = "ODEHingeJoint",
                 body1 = None,
                 body2 = None,
                 **params):
        ODEJointBase.__init__(self, name=name, body1=body1, body2=body2,
                              **params)

        self._createSlots()

        self.angle_slot = ProceduralDoubleSlot(self.computeAngle)
        self.addSlot("angle", self.angle_slot)
        getScene().timer().time_slot.addDependent(self.angle_slot)

#???
#    def reset(self):
#        if self.odejoint!=None:
#            print "ANGLE:",self.odejoint.getAngle(), self.angle
#            self.odejoint.setAnchor(self.odejoint.getAnchor())
        
    # computeAngle
    def computeAngle(self):
        """Return the current angle.

        This method is used as procedure for the angle_slot.
        """
        if self.odejoint==None:
            return 0.0
        else:
            return self.odejoint.getAngle()            

    # _createODEjoint
    def _createODEjoint(self):
        # Create the ODE joint
        self.odejoint = ode.HingeJoint(self.odedynamics.world)

    # _initODEjoint
    def _initODEjoint(self):
        ODEJointBase._initODEjoint(self)

        W = self.worldtransform
        p = W[3]
        self.odejoint.setAnchor((p.x, p.y, p.z))
        a = -W[0]
#        print "AXIS:",a,self.name
        self.odejoint.setAxis((a.x, a.y, a.z))
        self.odejoint.setParam(ode.ParamFMax, self.motorfmax)
        self.odejoint.setParam(ode.ParamVel, self.motorvel)



# FixedJoint
class ODEFixedJoint(ODEJointBase):
    """
    Fixed Joint: Glues two bodies together.
    Not recommended by ODE manual, but useful when a solid body has different contact properties on different sides.
    """

    def __init__(self,
                 name = "ODEFixedJoint",
                 body1 = None,
                 body2 = None,
                 **params):
        ODEJointBase.__init__(self, name=name, body1=body1, body2=body2,
                              **params)

        self._createSlots()

    # _createODEjoint
    def _createODEjoint(self):
        # Create the ODE joint
        self.odejoint = ode.FixedJoint(self.odedynamics.world)

    # _initODEjoint
    def _initODEjoint(self):
        ODEJointBase._initODEjoint(self)
        self.odejoint.setFixed()




# SliderJoint
class ODESliderJoint(ODEJointBase):
    """
    the slider axis is the local x axis
    """

    exec(slotPropertyCode("position"))

    def __init__(self,
                 name = "ODEHingeJoint",
                 body1 = None,
                 body2 = None,
                 **params):
        ODEJointBase.__init__(self, name=name, body1=body1, body2=body2,
                              **params)

        self._createSlots()

        self.position_slot = ProceduralDoubleSlot(self.computePosition)
        self.addSlot("position", self.position_slot)
        getScene().timer().time_slot.addDependent(self.position_slot)

    # computePosition
    def computePosition(self):
        """Return the current slider position.

        This method is used as procedure for the position_slot.
        """
        if self.odejoint==None:
            return 0.0
        else:
            return self.odejoint.getPosition()            

    # _createODEjoint
    def _createODEjoint(self):
        # Create the ODE joint
        self.odejoint = ode.SliderJoint(self.odedynamics.world)

    # _initODEjoint
    def _initODEjoint(self):
        ODEJointBase._initODEjoint(self)

        W = self.worldtransform
        a = W[0]
        self.odejoint.setAxis((a.x, a.y, a.z))
        self.odejoint.setParam(ode.ParamFMax, self.motorfmax)
        self.odejoint.setParam(ode.ParamVel, self.motorvel)


# Hinge2Joint
class ODEHinge2Joint(ODEJointBase):
    """
    axis1 is the local z axis
    """

    exec(slotPropertyCode("motorfmax"))
    exec(slotPropertyCode("motorvel"))
    exec(slotPropertyCode("suspensionerp"))
    exec(slotPropertyCode("suspensioncfm"))
    exec(slotPropertyCode("motorfmax2"))
    exec(slotPropertyCode("motorvel2"))

    def __init__(self,
                 name = "ODEHingeJoint",
                 body1 = None,
                 body2 = None,
                 **params):
        ODEJointBase.__init__(self, name=name, body1=body1, body2=body2,
                              **params)

        self._createSlots()
        self._createSlot("suspensionerp", None, "0.2", "onSuspensionERPChanged")
        self._createSlot("suspensioncfm", None, "1E-7", "onSuspensionCFMChanged")

        self._createSlot("motorvel2", None, "0.0", "onMotorVel2Changed")
        self._createSlot("motorfmax2", None, "0.0", "onMotorFMax2Changed")


        # default axis2 is the local y axis
        self.axis2_local = vec4(0,1,0,0)

    # _createODEjoint
    def _createODEjoint(self):
        # Create the ODE joint
        self.odejoint = ode.Hinge2Joint(self.odedynamics.world)

    # _initODEjoint
    def _initODEjoint(self):
        ODEJointBase._initODEjoint(self)

        W = self.worldtransform
        p = W[3]
        self.odejoint.setAnchor((p.x, p.y, p.z))
        a = W[2]
        self.odejoint.setAxis1((a.x, a.y, a.z))
        a = W*self.axis2_local
        self.odejoint.setAxis2((a.x, a.y, a.z))
        
        self.odejoint.setParam(ode.ParamFMax, self.motorfmax)
        self.odejoint.setParam(ode.ParamVel, self.motorvel)

    def onSuspensionERPChanged(self):
        print(("susp. erp has been changed to",self.suspensionerp_slot.getValue()))
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamSuspensionERP,
                                   self.suspensionerp_slot.getValue())

    def onSuspensionCFMChanged(self):
        print(("susp. cfm has been changed to",self.suspensioncfm_slot.getValue()))
        if self.odejoint!=None:
            self.odejoint.setParam(ode.ParamSuspensionCFM,
                                   self.suspensioncfm_slot.getValue())


# UniversalJoint
class ODEUniversalJoint(ODEJointBase):
    """

    """

    exec(slotPropertyCode("angle"))

    def __init__(self,
                 name = "ODEUniversalJoint",
                 body1 = None,
                 body2 = None,
                 **params):
        ODEJointBase.__init__(self, name=name, body1=body1, body2=body2,
                              **params)

        self._createSlots()

#        self.angle_slot = ProceduralDoubleSlot(self.computeAngle)
#        self.addSlot("angle", self.angle_slot)
#        getScene().timer().time_slot.addDependent(self.angle_slot)


    # _createODEjoint
    def _createODEjoint(self):
        # Create the ODE joint
        self.odejoint = ode.UniversalJoint(self.odedynamics.world)

    # _initODEjoint
    def _initODEjoint(self):
        ODEJointBase._initODEjoint(self)

        W = self.worldtransform
        p = W[3]
        self.odejoint.setAnchor((p.x, p.y, p.z))
        a = W[0]
        self.odejoint.setAxis1((a.x, a.y, a.z))
        a = W[1]
        self.odejoint.setAxis2((a.x, a.y, a.z))
        self.odejoint.setParam(ode.ParamFMax, self.motorfmax)
        self.odejoint.setParam(ode.ParamVel, self.motorvel)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: bvh.py,v 1.1 2005/02/06 22:26:02 mbaas Exp $

## \file bvh.py
## Contains the BVHReader class.

import string

# Node
class Node:
    def __init__(self, root=False):
        self.name = None
        self.channels = []
        self.offset = (0,0,0)
        self.children = []
        self._is_root = root

    def isRoot(self):
        return self._is_root

    def isEndSite(self):
        return len(self.children)==0
    

# BVHReader
class BVHReader:
    """Read BioVision Hierarchical (BVH) files.
    """

    def __init__(self, filename):

        self.filename = filename
        # A list of unprocessed tokens (strings)
        self.tokenlist = []
        # The current line number
        self.linenr = 0

        # Root node
        self._root = None
        self._nodestack = []

        # Total number of channels
        self._numchannels = 0

    def onHierarchy(self, root):
        pass

    def onMotion(self, frames, dt):
        pass

    def onFrame(self, values):
        pass

    # read
    def read(self):
        """Read the entire file.
        """
        self.fhandle = file(self.filename)

        self.readHierarchy()
        self.onHierarchy(self._root)
        self.readMotion()

    # readMotion
    def readMotion(self):
        """Read the motion samples.
        """
        # No more tokens (i.e. end of file)? Then just return 
        try:
            tok = self.token()
        except StopIteration:
            return
        
        if tok!="MOTION":
            raise SyntaxError("Syntax error in line %d: 'MOTION' expected, got '%s' instead"%(self.linenr, tok))

        # Read the number of frames
        tok = self.token()
        if tok!="Frames:":
            raise SyntaxError("Syntax error in line %d: 'Frames:' expected, got '%s' instead"%(self.linenr, tok))

        frames = self.intToken()

        # Read the frame time
        tok = self.token()
        if tok!="Frame":
            raise SyntaxError("Syntax error in line %d: 'Frame Time:' expected, got '%s' instead"%(self.linenr, tok))
        tok = self.token()
        if tok!="Time:":
            raise SyntaxError("Syntax error in line %d: 'Frame Time:' expected, got 'Frame %s' instead"%(self.linenr, tok))

        dt = self.floatToken()

        self.onMotion(frames, dt)

        # Read the channel values
        for i in range(frames):
            s = self.readLine()
            a = s.split()
            if len(a)!=self._numchannels:
                raise SyntaxError("Syntax error in line %d: %d float values expected, got %d instead"%(self.linenr, self._numchannels, len(a)))
            values = [float(x) for x in a]
            self.onFrame(values)


    # readHierarchy
    def readHierarchy(self):
        """Read the skeleton hierarchy.
        """
        tok = self.token()
        if tok!="HIERARCHY":
            raise SyntaxError("Syntax error in line %d: 'HIERARCHY' expected, got '%s' instead"%(self.linenr, tok))

        tok = self.token()
        if tok!="ROOT":
            raise SyntaxError("Syntax error in line %d: 'ROOT' expected, got '%s' instead"%(self.linenr, tok))

        self._root = Node(root=True)
        self._nodestack.append(self._root)
        self.readNode()

    # readNode
    def readNode(self):
        """Read the data for a node.
        """

        # Read the node name (or the word 'Site' if it was a 'End Site'
        # node)
        name = self.token()
        self._nodestack[-1].name = name
        
        tok = self.token()
        if tok!="{":
            raise SyntaxError("Syntax error in line %d: '{' expected, got '%s' instead"%(self.linenr, tok))

        while 1:
            tok = self.token()
            if tok=="OFFSET":
                x = self.floatToken()
                y = self.floatToken()
                z = self.floatToken()
                self._nodestack[-1].offset = (x,y,z)
            elif tok=="CHANNELS":
                n = self.intToken()
                channels = []
                for i in range(n):
                    tok = self.token()
                    if tok not in ["Xposition", "Yposition", "Zposition",
                                  "Xrotation", "Yrotation", "Zrotation"]:
                        raise SyntaxError("Syntax error in line %d: Invalid channel name: '%s'"%(self.linenr, tok))                        
                    channels.append(tok)
                self._numchannels += len(channels)
                self._nodestack[-1].channels = channels
            elif tok=="JOINT":
                node = Node()
                self._nodestack[-1].children.append(node)
                self._nodestack.append(node)
                self.readNode()
            elif tok=="End":
                node = Node()
                self._nodestack[-1].children.append(node)
                self._nodestack.append(node)
                self.readNode()
            elif tok=="}":
                if self._nodestack[-1].isEndSite():
                    self._nodestack[-1].name = "End Site"
                self._nodestack.pop()
                break
            else:
                raise SyntaxError("Syntax error in line %d: Unknown keyword '%s'"%(self.linenr, tok))
        

    # intToken
    def intToken(self):
        """Return the next token which must be an int.
        """

        tok = self.token()
        try:
            return int(tok)
        except ValueError:
            raise SyntaxError("Syntax error in line %d: Integer expected, got '%s' instead"%(self.linenr, tok))

    # floatToken
    def floatToken(self):
        """Return the next token which must be a float.
        """

        tok = self.token()
        try:
            return float(tok)
        except ValueError:
            raise SyntaxError("Syntax error in line %d: Float expected, got '%s' instead"%(self.linenr, tok))

    # token
    def token(self):
        """Return the next token."""

        # Are there still some tokens left? then just return the next one
        if self.tokenlist!=[]:
            tok = self.tokenlist[0]
            self.tokenlist = self.tokenlist[1:]
            return tok

        # Read a new line
        s = self.readLine()
        self.createTokens(s)
        return self.token()

    # readLine
    def readLine(self):
        """Return the next line.

        Empty lines are skipped. If the end of the file has been
        reached, a StopIteration exception is thrown.  The return
        value is the next line containing data (this will never be an
        empty string).
        """
        # Discard any remaining tokens
        self.tokenlist = []
      
        # Read the next line
        while 1:
            s = self.fhandle.readline()
            self.linenr += 1
            if s=="":
                raise StopIteration
            return s

    # createTokens
    def createTokens(self, s):
        """Populate the token list from the content of s.
        """
        s = s.strip()
        a = s.split()
        self.tokenlist = a

    



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: irenderer.py,v 1.1.1.1 2004/12/12 14:31:42 mbaas Exp $

class IRenderer:
    """The basic renderer interface.
    
    The renderer is responsible for displaying the scene on the screen
    in real time (i.e. frame rates that allow working with the scene
    interactively). 
    
    There are actually two classes involved in rendering, the Renderer
    (this class) and a RenderInstance class (IRenderInstance). The
    Renderer class represents a particlar rendering algorithm or
    render engine while the RenderInstance is used for rendering on
    a particular widget. All RenderInstances that were spawned from
    the same Renderer can share resources like display lists, textures, etc.
        
    Before a scene can be rendered a renderer object has to be created
    that is connected to one particular scene (which is passed to the 
    constructor). Then, for each view you create a RenderInstance object
    by calling createRenderInstance().
    On a render instance you can set various parameters like render flags, 
    camera and viewport parameters via 
    \link IRenderInstance::setRenderFlags setRenderFlags() \endlink, 
    \link IRenderInstance::setCamera setCamera() \endlink and
    \link IRenderInstance::setViewport setViewport() \endlink. Once 
    everything is set up, the scene can be
    rendered by calling \link IRenderInstance::paint paint() \endlink for 
    every frame. A picking operation can be performed using the 
    \link IRenderInstance::pick pick() \endlink method.
    
    Of course, an actual implementation of the interface may actually
    contain more methods to activate special features supported by
    this particular renderer (e.g. shadows, shaders, ...).
    
    There are two "flavors" of renderers, one that expects the application
    to provide the rendering widget and one that creates its own window.
    The interface is the same, only the semantics of some methods change
    slightly. If a renderer creates its own window it has to do so in
    the createRenderInstance() method and close the window again in 
    \link IRenderInstance::close RenderInstance.close() \endlink. 
    An embedded renderer can assume that its 
    associated rendering context was previously made current by the 
    application for most of the calls (as noted in the
    method documentations).
    
    \see IRenderInstance
    """
        
    def __init__(self, scene):
        """Constructor.
        
        \param scene (\c IScene) The scene interface of the scene that
            is to be rendered.
        """
        
    def createRenderInstance(self):
        """Open & initialize a render instance.
        
        The renderer will set default values for the flags, camera
        and viewport.
        
        If the renderer is embedded this method may only be called if
        the associated rendering context (OpenGL context) is made current.        

        \return Render instance object (\c RenderInstance)
        """
        
        
        
class IRenderInstance:
    """The render instance.
    
    \see IRenderer
    """

    def close(self):
        """Close the renderer interface.
        
        Frees any resources that were previously allocated.
        """

    def setRenderFlags(self, flags):
        """Set the flags that determine how the scene is rendered.
        
        The \a flags parameter is a combination of the following flags:
            
        - \c SOLID: Draw the scene using shaded solid faces
        - \c WIREFRAME: Draw the scene using unlit wireframes (this flag can 
             be combined with SOLID)
        - \c SMOOTH: Use smooth shading, otherwise flat (has only an effect
             when SOLID is set)
        - \c TEXTURE: Use texture mapping (has only an effect when SOLID
             is set)
        
        \param flags (\c int) Flags
        """
        
    def getRenderFlags(self):
        """Return the current render flags.
        
        \return Flags (\c int).
        \see setRenderFlags()
        """
        
    def setCamera(self, cam):
        """Set the camera to use for viewing the scene.
        
        The camera interface \a cam is used to retrieve the location
        and orientation of the camera (= inverse view transformation)
        and the projection matrix (which determines if the camera
        is an orthographic or perspective camera).
        
        \param cam (\c IViewCamera) Camera interface
        """
        
    def getCamera(self):
        """Return the currently active camera.
        
        \return Camera (\c IViewCamera)
        \see setCamera()
        """
        
    def setViewport(self, viewport):
        """Set the viewport settings.
        
        \a viewport contains the position and size (x,y,width,height) of 
        the viewing area in pixels.
        
        If the renderer is embedded this method may only be called if
        the associated rendering context (OpenGL context) is made current.
        
        \param viewport (\c IViewportSettings) Viewport settings
        """
        
    def getViewport(self):
        """Return the viewport settings.
        
        \return Viewport settings (\c IViewportSettings).
        \see setViewport()
        """
    
    def paint(self):
        """Renders the scene.
        
        The scene that was given to the open() call is rendered
        using the current camera, viewport settings and render flags.
        
        If the renderer is embedded this method may only be called if
        the associated rendering context (OpenGL context) is made current.
        
        \see setRenderFlags(), setCamera(), setViewport()
        """
        
    def pick(self, x, y):
        """Performs a picking operation and returns the picked objects.
        
        Performs a picking operation at pixel position (\a x, \a y)
        and returns all objects that were hit in depth sorted order (nearest
        first).
        
        If the renderer is embedded this method may only be called if
        the associated rendering context (OpenGL context) is made current.
                
        \param x (\c int) Pick position in pixels (X)
        \param y (\c int) Pick position in pixels (Y)
        """
        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: tunnel.py,v 1.2 2005/08/08 18:05:34 mbaas Exp $

## \file tunnel.py
## Contains the Tunnel class.

from .cgtypes import *
from . import component
from .slots import *
from . import _core
import socket, threading, struct

# _ImmediateForwarder
class _ImmediateForwarder(NotificationForwarder):
    """Internal class for the Tunnel component.

    This forwarder immediately sends the new value to the server part
    of the tunnel.
    """
    def __init__(self, tunnel, slot, msgid, idx):
        NotificationForwarder.__init__(self, self.sendValue)
        self.tunnel = tunnel
        self.slot = slot
        self.msgid = msgid
        self.idx = idx

    def sendValue(self):
        t = self.tunnel
        msg = t.encodeValueMessage(self.msgid, self.idx, self.slot)
        t.send(msg)

# _GatherForwarder
class _GatherForwarder(NotificationForwarder):
    """Internal class for the Tunnel component.

    This forwarder marks the slot as changed and sends a combined
    message if it's the sender slot (the last slot in the tunnel).
    """
    def __init__(self, tunnel, idx, sender):
        NotificationForwarder.__init__(self, self.markSlot)
        self.tunnel = tunnel
        self.idx = idx
        self.sender = sender

    def markSlot(self):
        self.tunnel.markAsChanged(self.idx)
        if self.sender:
            self.tunnel.sendBulk()

                 

# Tunnel
class Tunnel(component.Component):
    """%Tunnel component which can connect slots on different machines.

    A tunnel is a component that has an arbitrary number of input slots
    and a corresponding output slot for each input slot.
    The value of the input slot is simply passed to the output slot.
    The speciality of the tunnel is that both ends may reside on two
    different machines.

    \image html tunnel.png

    An instance of the tunnel component either represents the input part
    (client) or the output part (server). Value changes on the client are
    then propagated to the server. When creating either side of the tunnel
    you have to define the name and type of the slots that should be created.
    The slot types of the client and the server should always match.

    Note: ArraySlots are currently not supported.

    Example:

    \code
    # On machine A
    t = Tunnel(
       slots = [("pos", "Vec3Slot()"),
                ("rot", "Mat3Slot()")],
       host  = "<name or IP of machine B>",
    )

    # Do connections between other components and t.pos_slot/t.rot_slot
    # or set the values directly on the tunnel
    \endcode

    \code
    # On machine B
    t = Tunnel(
       server = True,
       slots = [("pos", "Vec3Slot()"),
                ("rot", "Mat3Slot()")],
    )
    
    # Make connections between t.pos_slot/t.rot_slot and other components
    # or retrieve the values directly from the tunnel
    \endcode

    \par Protocol

    The client sends its messages as UDP datagrams to the server.
    One packet may contain several individual messages that are just
    concatenated. The first two bytes of each message is the message
    ID. The remaining part depends on the ID. The byte order is always
    in big endian.

    - ID 0: Init message (must be the only message in a packet)
    - ID 1: Binary int value
    - ID 2: Binary double value
    - ID 3: Binary bool value (actually short int)
    - ID 4: Binary vec3 value (3 doubles)
    - ID 5: Binary vec4 value (4 doubles)
    - ID 6: Binary mat3 value (9 doubles)
    - ID 7: Binary mat4 value (16 doubles)
    - ID 8: Binary quat value (4 doubles)

    All the value messages are composed of the slot index number (short int),
    followed by the binary value.    
    """

    def __init__(self,
                 name = "Tunnel",
                 server = False,
                 slots = None,
                 port = 64738,
                 host = "localhost",
                 gather_messages = False,
                 verbose = False,
                 auto_insert = True):
        """Constructor.

        The argument \a slot specifies the slots to create on the tunnel.
        It must be a list of 2-tuples containing two strings. The first
        string is the name of the attribute and the second contains the
        type and initializer of the slot (in Python syntax). The order and
        types of the slots on the client and server should always match.

        The arguments \a host and \a gather_messages are only meaningful
        on the client side. If \a gather_messages is True, then value changes
        on a slot are not immediately sent but only when the last slot
        receives a new value. You can use this option if you know that
        all slot values will change at the same time. In this case it's more
        efficient to change all values and then send only one message
        containing all new values instead of sending individual messages for
        every slot.

        \param name (\c str) Component name
        \param server (\c bool) True if this is the server part
        \param slots (\c list) A list of slot definitions. Each definition is a tuple (attribute name, slot class and initial arguments).
        \param port (\c int) Port number to use for the data transfer
        \param host (\c str) Host where the tunnel server is running (name or IP address). Client only.
        \param gather_messages (\c bool) True if the messages should be combined. Client only.
        \param verbose (\c bool) True if debug messages should be printed
        \param auto_insert (\c bool) True if the component should be inserted into the scene automatically
        """
        component.Component.__init__(self, name=name, auto_insert=auto_insert)

        if slots==None:
            slots = []

        self.server = server
        self.host = host
        self.port = port
        self.sock = None
        self.slots = slots
        self.gather_messages = gather_messages
        self.verbose = verbose

        if self.server:
            self.host = socket.gethostname()
            
        # A list with tuples (slot, id)
        self.slotobjs = []
        # A dictionary with slot indices as keys. If an index is present
        # in a dictionary then the corresponding slot was modified and the
        # value has to be sent to the server
        self.changed = {}

        self.forwarders = []

        # Initialize client/server...
        if server:
            self.initServer()
        else:
            self.initClient()

    def __str__(self):
        if self.server:
            return 'Tunnel server "%s" listening at %s:%d. %d slots.'%(self.name, self.host, self.port, len(self.slots))
        else:
            return 'Tunnel client "%s". %d slots. Server at %s:%d.'%(self.name, len(self.slots), self.host, self.port)

    ## protected:

    # initClient
    def initClient(self):
        """Initialization method for the client.
        """
        # Create socket object for sending UDP datagrams
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        # Create slot attributes...
        self.createClientSlotAttribs()

        # Send init message
        s = ["%s:%s"%x for x in self.slots]
        s = ";".join(s)
        self.send(struct.pack(">H", 0)+s)

    # initServer
    def initServer(self):
        """Initialization method for the server.
        """
        # Create slot attributes...
        self.createServerSlotAttribs()
        
        # Open UDP socket
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        ipaddr = socket.gethostbyname(self.host)
        if self.verbose:
            print(("Open UDP port %d on %s (%s)"%(self.port, ipaddr, self.host)))
        self.sock.bind((ipaddr, self.port))

        # Split thread with the server loop
        self.serverthread = threading.Thread(name="Tunnel-Server",
                                             target=self.serverLoop)
        self.serverthread.setDaemon(True)
        self.serverthread.start()    

    # serverLoop
    def serverLoop(self):
        """Server loop.

        This method reads messages from the previously created socket
        and process them.
        The method is run in its own thread (because reading from the
        socket is a blocking operation).
        """

        if self.verbose:
            print(("Tunnel server running at port",self.port))
        while 1:
            # Read a packet (blocking)
            rawdata, addr = self.sock.recvfrom(5000)
            if self.verbose:
                print("---Tunnel messages--")
            # Process all messages...
            while rawdata!="":
                # Get the id of the first message...
                try:
                    id = struct.unpack(">H", rawdata[:2])[0]
                except:
                    print("Error: Invalid UDP packet (no valid message id).")
                    rawdata = ""
                    continue

                # Process the message...
                msgdata = rawdata[2:]
                if id==0:
                    if self.verbose:
                        print("Init")
                    f = msgdata.split(";")
                    localslots = [x[1] for x in list(self.iterSlotDefs(self.slots))]
                    remoteslots = [tuple(x.split(":")) for x in f]
                    remoteslots = [x[1] for x in list(self.iterSlotDefs(remoteslots))]
                    mismatch = True
                    if len(localslots)==len(remoteslots):
                        mismatch = localslots!=remoteslots
                    if mismatch:
                        print(("%s: The types of the remote and local slots don't match."%self.name))
                    rawdata = ""
                else:
                    # Decode the slot index and the value to set...
                    try:
                        n,idx,v = self.decodeValueMessage(id, msgdata)
                    except struct.error as e:
                        print(e)
                        rawdata = ""
                        continue
                    except ValueError as e:
                        print(e)
                        rawdata = ""
                        continue
                    # Prepare the next message...
                    rawdata = msgdata[n:]
                    if self.verbose:
                        print(("Message id: %d - Slot:%d Value:%s"%(id,idx,v)))
                    # Get the slot that will receive the new value...
                    try:
                        slot,id = self.slotobjs[idx]
                    except IndexError as e:
                        print(("Error: Invalid slot index (%d)"%idx))
                    # Set the new value...
                    try:
                        slot.setValue(v)
                    except:
                        print(("Error: Could not assign value",v,"to slot of type",slot.typeName()))

        self.sock.close()
        if self.verbose:
            print(("Tunnel server stopped at port",self.port))

    # markAsChanged
    def markAsChanged(self, idx):
        """Mark a slot that has changed its value.
        """
        self.changed[idx] = 1

    # send
    def send(self, msg):
        """Send one or more messages to the server.

        \param msg (\c str) Message(s)
        """
        self.sock.sendto(msg, (self.host, self.port))

    # sendBuld
    def sendBulk(self):
        """Send the values of all slots that have changed since the last call.
        """
        msgs = ""
        for idx in list(self.changed.keys()):
            slot,id = self.slotobjs[idx]
            msgs += self.encodeValueMessage(id, idx, slot)

        self.send(msgs)
        self.changed = {}

    # createClientSlotAttribs
    def createClientSlotAttribs(self):
        slots = self.slots
        if slots==None:
            return

        self.forwarders = []
        slotnames = ["<none>", "IntSlot", "DoubleSlot", "BoolSlot", "Vec3Slot",
                     "Vec4Slot", "Mat3Slot", "Mat4Slot", "QuatSlot"]

        for varname, slotname, slotparams in self.iterSlotDefs(slots):
            # Create the slot object
            exec("slot = %s%s"%(slotname, slotparams))
            # Add the slot as attribute
            setattr(self, "%s_slot"%varname, slot)
            exec("self.addSlot('%s', self.%s_slot)"%(varname, varname))

            id = slotnames.index(slotname)
            idx = len(self.forwarders)
            
            self.slotobjs.append((slot,id))

            # Create the forwarder object
            if self.gather_messages:
                f = _GatherForwarder(self, idx, idx==len(slots)-1)
            else:
                f = _ImmediateForwarder(self, slot, id, idx)
            slot.addDependent(f)
            self.forwarders.append(f)

    # createServerSlotAttribs
    def createServerSlotAttribs(self):
        slots = self.slots
        if slots==None:
            return

        slotnames = ["<none>", "IntSlot", "DoubleSlot", "BoolSlot", "Vec3Slot",
                     "Vec4Slot", "Mat3Slot", "Mat4Slot", "QuatSlot"]

        for varname, slotname, slotparams in self.iterSlotDefs(slots):
            # Create the slot object
            exec("slot = %s%s"%(slotname, slotparams))
            # Add the slot as attribute
            setattr(self, "%s_slot"%varname, slot)
            exec("self.addSlot('%s', self.%s_slot)"%(varname, varname))

            id = slotnames.index(slotname)
            idx = len(self.forwarders)
            
            self.slotobjs.append((slot,id))


    # deleteSlotAttribs
    def deleteSlotAttribs(self):
        if self.slots==None:
            return
        
        for name,slot in self.slots:
            slotname = "%s_slot"%name
            exec("del self.%s"%slotname)

        self.slots = None

    # encodeValueMessage
    def encodeValueMessage(self, id, idx, slot):
        """Encode a binary value message.

        \param id (\c int) Message ID
        \param idx (\c int) Slot index
        \param slot (\c Slot) Corresponding slot
        """

        # int
        if id==1:
            return struct.pack(">HHi", id, idx, slot.getValue())
        # double
        elif id==2:
            return struct.pack(">HHd", id, idx, slot.getValue())
        # bool
        elif id==3:
            return struct.pack(">HHh", id, idx, slot.getValue())
        # vec3
        elif id==4:
            x,y,z = slot.getValue()
            return struct.pack(">HHddd", id, idx, x, y, z)
        # vec4
        elif id==5:
            x,y,z,w = slot.getValue()
            return struct.pack(">HHdddd", id, idx, x, y, z,w)
        # mat3
        elif id==6:
            M = slot.getValue()
            return struct.pack(">HHddddddddd", id, idx, *M.toList(rowmajor=True))
        # mat4
        elif id==7:
            M = slot.getValue()
            return struct.pack(">HHdddddddddddddddd", id, idx, *M.toList(rowmajor=True))
        # quat
        elif id==8:
            q = slot.getValue()
            return struct.pack(">HHdddd", id, idx, q.w, q.x, q.y, q.z)

    # decodeValueMessage
    def decodeValueMessage(self, id, msg):
        """Decode a binary value message.

        Raises a ValueError exception if the id is invalid. If the
        size of msg is too short, a struct.error exception is thrown.

        \param id (\c int) Message ID
        \param msg (\c str) Data part of the message (may contain additional messages)
        \return Tuple (Message size, slot index, value)
        """

        # int
        if id==1:
            id, v = struct.unpack(">Hi", msg[:6])
            return 6,id,v
        # double
        elif id==2:
            id, v = struct.unpack(">Hd", msg[0:10])
            return 10,id,v
        # bool
        elif id==3:
            id, v = struct.unpack(">Hh", msg[:4])
            return 4,id,v
        # vec3
        elif id==4:
            idx, x,y,z = struct.unpack(">Hddd", msg[:26])
            return (26, idx, vec3(x,y,z))
        # vec4
        elif id==5:
            idx, x,y,z,w = struct.unpack(">Hdddd", msg[:34])
            return (34, idx, vec4(x,y,z,w))
        # mat3
        elif id==6:
            idx, m11,m12,m13,m21,m22,m23,m31,m32,m33 = struct.unpack(">Hddddddddd", msg[:74])
            return (74, idx, mat3(m11,m12,m13,m21,m22,m23,m31,m32,m33))
        # mat4
        elif id==7:
            idx, m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = struct.unpack(">Hdddddddddddddddd", msg[:130])
            return (130, idx, mat4(m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44))
        # quat
        elif id==8:
            idx, w,x,y,z = struct.unpack(">Hdddd", msg[:34])
            return (34, idx, quat(w,x,y,z))
        else:
            raise ValueError("Unknown message id (%d)"%id)
            

    # iterSlotDefs
    def iterSlotDefs(self, slots):
        """Generates the variable name, slot name and slot parameters.

        This generator method takes a slot definition list and generates
        3-tuples (variable name, slot name, slot parameters).

        For example, the slot definition list [("xpos", "IntSlot(2)"),
        ("ypos", "IntSlot(12)")] will generate two tuples ("xpos", "IntSlot",
        "(2)") and ("ypos", "IntSlot", "(12)").
        """
        
        for varname, slotdef in slots:
            n = slotdef.find("(")
            if n!=-1:
                slotname = slotdef[:n]
                slotparams = slotdef[n:]
            else:
                slotname = slotdef
                slotparams = "()"
            yield varname, slotname, slotparams
        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2008
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# -------------------------------------------------------------
# The RenderMan (R) Interface Procedures and Protocol are:
# Copyright 1988, 1989, 2000, Pixar
# All Rights Reserved
#
# RenderMan (R) is a registered trademark of Pixar
# -------------------------------------------------------------
# $Id: ri.py,v 1.4 2006/02/14 19:29:39 mbaas Exp $

"""This is the "lower-level" _cri module which is used by the cri module.

This module provides the function loadRI() which loads a shared RenderMan
library and prepares the returned library handle, so that all RenderMan
constants, tokens and functions are properly defined.

The ri library returned by loadRI() has to be used like a C library and
can crash the application if not used properly (like forgetting to terminate
parameter lists with RI_NULL). This is why this low-level ri library is not
passed back to the user but instead the cri module provides another wrapper
layer that allows the library to be used just like the pure Python ri module.
"""

import os.path
from ctypes import *
from . import rmanlibutil

def loadRI(libName):
    """Load a RenderMan library and return a module-like handle to it.
    
    libName is the name of a shared library that implements the RenderMan
    interface. The name can either be an absolute file name or just the
    name of the library (without suffix or "lib" prefix) in which case 
    the function tries to find the library file itself.
    The return value is the library handle as returned by the ctypes 
    LoadLibrary() function. This handle has already been prepared so that
    it can be used like a module that contains the RenderMan interface.
    """
    
    # Try to figure out the location of the lib if the name is not an absolute path...
    libName = rmanlibutil.resolveRManLib(libName)
        
    # Load the library...
    ri = cdll.LoadLibrary(libName)
    
    _createRiTypes(ri)
    _createRiTokens(ri)
    _createRiConstants(ri)
    _createRiFunctions(ri)

    # Create an alias for every Ri function and RI_ constant that has the prefix removed...
    for name in dir(ri):
        if name.startswith("Ri"):
            setattr(ri, name[2:], getattr(ri, name))
        elif name.startswith("RI_"):
            setattr(ri, name[3:], getattr(ri, name))

    ri.__class__.RiLastError = property(_getLastError, _setLastError)
    ri.__class__.LastError = property(_getLastError, _setLastError)

    return ri

def importRINames(ri, ns):
    """Import the RenderMan names into the given namespace.
    """
    for name in dir(ri):
        if name.startswith("Ri"):
            localName = name
            ns[localName] = getattr(ri, name)
            
        if name[:2] in ["Rt", "RI"]:
            ns[name] = getattr(ri, name)


def _createRiTypes(ri):
    """Create the RenderMan types.
    
    ri must be the open ctypes library handle.
    The RenderMan types are added as attributes to the ri object. All names
    begin with "Rt" (RtInt, RtFloat, ...).
    
    This is a helper function for loadRI().
    """
    
    # Base types that are not composed of other RenderMan types...
    baseTypes = dict(RtBoolean = c_short,
                     RtInt = c_int,
                     RtFloat = c_float,
                     RtString = c_char_p,
                     RtToken = c_char_p,
                     RtVoid = None,
                     RtPointer = c_void_p)
    
    for name,ctype in list(baseTypes.items()):
        setattr(ri, name, ctype)
        
    ri.RtColor = 3*ri.RtFloat
    ri.RtPoint = 3*ri.RtFloat
    ri.RtVector = 3*ri.RtFloat
    ri.RtNormal = 3*ri.RtFloat
    ri.RtHpoint = 4*ri.RtFloat
    ri.RtMatrix = 16*ri.RtFloat
    ri.RtBasis = 16*ri.RtFloat
    ri.RtBound = 6*ri.RtFloat
    
    ri.RtObjectHandle = ri.RtPointer
    ri.RtLightHandle = ri.RtPointer
    ri.RtContextHandle = ri.RtPointer
    
    ri.RtFilterFunc = CFUNCTYPE(ri.RtFloat,  ri.RtFloat, ri.RtFloat, ri.RtFloat, ri.RtFloat)
    ri.RtErrorHandler = CFUNCTYPE(ri.RtVoid,  ri.RtInt, ri.RtInt, c_char_p)
    ri.RtProcSubdivFunc = CFUNCTYPE(ri.RtVoid,  ri.RtPointer, ri.RtFloat)
    ri.RtProcFreeFunc = CFUNCTYPE(ri.RtVoid,  ri.RtPointer)
    ri.RtArchiveCallback = CFUNCTYPE(ri.RtVoid,  ri.RtToken, c_char_p)   # var args are missing
    
    
def _createRiConstants(ri):
    """Create the RenderMan constants.
    
    Add the RenderMan constants to the ri object which must be an open
    ctypes library handle.
    The types must already be available on ri (i.e. _createRiTypes() must
    already have been called).
    
    This is a helper function for loadRI().
    """
    
    ri.RI_NULL         = None
    ri.RI_TRUE         = 1
    ri.RI_FALSE        = 0
    ri.RI_EPSILON      = 1.0e-10
    ri.RI_INFINITY     = ri.RtFloat(1.0e38).value

    ri.RI_HERMITESTEP    = 2
    ri.RI_CATMULLROMSTEP = 1
    ri.RI_BEZIERSTEP     = 3
    ri.RI_BSPLINESTEP    = 1
    ri.RI_POWERSTEP      = 4

    bases = ["RiBezierBasis", "RiBSplineBasis", "RiCatmullRomBasis",
             "RiHermiteBasis", "RiPowerBasis"]
    
    for basis in bases:
        try:
            value = ri.RtBasis.in_dll(ri, basis)
        except ValueError:
            raise ValueError('The RenderMan implementation "%s" does not define the standard basis "%s"'%(getattr(ri, "_name", "?"), basis))
        setattr(ri, basis, value)
    
def _createRiTokens(ri):
    """Create the RenderMan tokens.
    
    The RenderMan constants are added as attributes to the ri object which
    must be an open ctypes library handle. All names begin with "RI_".
    
    This is a helper function for loadRI().
    """

    # All the following constants will be added to the ri object.
    # The items are (Token Name, Default). If a token does not exist in the
    # library, the default value is used.
    tokens = [("RI_A", "a"),
              ("RI_ABORT", "abort"),
              ("RI_AMBIENTLIGHT", "ambientlight"),
              ("RI_AMPLITUDE", "amplitude"),
              ("RI_AZ", "az"),
              ("RI_BACKGROUND", "background"),
              ("RI_BEAMDISTRIBUTION", "beamdistribution"),
              ("RI_BICUBIC", "bicubic"),
              ("RI_BILINEAR", "bilinear"),
              ("RI_BLACK", "black"),
              ("RI_BUMPY", "bumpy"),
              ("RI_CAMERA", "camera"),
              ("RI_CI", "Ci"),
              ("RI_CLAMP", "clamp"),
              ("RI_COMMENT", "comment"),
              ("RI_CONEANGLE", "coneangle"),
              ("RI_CONEDELTAANGLE", "conedeltaangle"),
              ("RI_CONSTANT", "constant"),
              ("RI_CONSTANTWIDTH", "constantwidth"),
              ("RI_CS", "Cs"),
              ("RI_CUBIC", "cubic"),
              ("RI_DEPTHCUE", "depthcue"),
              ("RI_DIFFERENCE", "difference"),
              ("RI_DISTANCE", "distance"),
              ("RI_DISTANTLIGHT", "distantlight"),
              ("RI_FILE", "file"),
              ("RI_FLATNESS", "flatness"),
              ("RI_FOG", "fog"),
              ("RI_FOV", "fov"),
              ("RI_FRAMEBUFFER", "framebuffer"),
              ("RI_FROM", "from"),
              ("RI_HANDLEID", "__handleid"),
              ("RI_HANDLER", "handler"),
              ("RI_HIDDEN", "hidden"),
              ("RI_IDENTIFIER", "identifier"),
              ("RI_IGNORE", "ignore"),
              ("RI_INSIDE", "inside"),
              ("RI_INTENSITY", "intensity"),
              ("RI_INTERSECTION", "intersection"),
              ("RI_KA", "Ka"),
              ("RI_KD", "Kd"),
              ("RI_KR", "Kr"),
              ("RI_KS", "Ks"),
              ("RI_LH", "lh"),
              ("RI_LIGHTCOLOR", "lightcolor"),
              ("RI_LINEAR", "linear"),
              ("RI_MATTE", "matte"),
              ("RI_MAXDISTANCE", "maxdistance"),
              ("RI_METAL", "metal"),
              ("RI_MINDISTANCE", "mindistance"),
              ("RI_N", "N"),
              ("RI_NAME", "name"),
              ("RI_NG", "Ng"),
              ("RI_NONPERIODIC", "nonperiodic"),
              ("RI_NP", "Np"),
              ("RI_OBJECT", "object"),
              ("RI_OI", "Oi"),
              ("RI_ORIGIN", "origin"),
              ("RI_ORTHOGRAPHIC", "orthographic"),
              ("RI_OS", "Os"),
              ("RI_OUTSIDE", "outside"),
              ("RI_P", "P"),
              ("RI_PAINT", "paint"),
              ("RI_PAINTEDPLASTIC", "paintedplastic"),
              ("RI_PERIODIC", "periodic"),
              ("RI_PERSPECTIVE", "perspective"),
              ("RI_PLASTIC", "plastic"),
              ("RI_POINTLIGHT", "pointlight"),
              ("RI_PRIMITIVE", "primitive"),
              ("RI_PRINT", "print"),
              ("RI_PW", "Pw"),
              ("RI_PZ", "Pz"),
              ("RI_RASTER", "raster"),
              ("RI_RGB", "rgb"),
              ("RI_RGBA", "rgba"),
              ("RI_RGBAZ", "rgbaz"),
              ("RI_RGBZ", "rgbz"),
              ("RI_RH", "rh"),
              ("RI_ROUGHNESS", "roughness"),
              ("RI_S", "s"),
              ("RI_SCREEN", "screen"),
              ("RI_SHADINGGROUP", "shadinggroup"),
              ("RI_SHINYMETAL", "shinymetal"),
              ("RI_SMOOTH", "smooth"),
              ("RI_SPECULARCOLOR", "specularcolor"),
              ("RI_SPOTLIGHT", "spotlight"),
              ("RI_ST", "st"),
              ("RI_STRUCTURE", "structure"),
              ("RI_T", "t"),
              ("RI_TEXTURENAME", "texturename"),
              ("RI_TO", "to"),
              ("RI_UNION", "union"),
              ("RI_VERBATIM", "verbatim"),
              ("RI_WIDTH", "width"),
              ("RI_WORLD", "world"),
              ("RI_Z", "z"),
              ]
    
    for tokSpec in tokens:
        if type(tokSpec) is str:
            name = tokSpec
            default = None
        else:
            try:
                name,default = tokSpec
                if type(name) is not str:
                    raise ValueError()
            except:
                raise ValueError("Invalid token spec: %s"%(tokSpec,))
            
        try:
            value = c_char_p.in_dll(ri, name).value
            # Turn into a unicode string
            value = value.decode("ascii")
        except ValueError:
            value = default
            
        setattr(ri, name, value)

        
def _createRiFunctions(ri):
    """Declare the RenderMan functions.
    
    ri must be an open ctypes library handle.
    
    This is a helper function for loadRI().
    """
    RtToken = ri.RtToken
    RtBoolean = ri.RtBoolean
    RtInt = ri.RtInt
    RtFloat = ri.RtFloat
    RtString = ri.RtString
    RtMatrix = ri.RtMatrix
    RtPoint = ri.RtPoint
    RtColor = ri.RtColor
    RtBasis = ri.RtBasis
    RtBound = ri.RtBound
    RtContextHandle = ri.RtContextHandle
    RtLightHandle = ri.RtLightHandle
    RtObjectHandle = ri.RtObjectHandle
    RtPointer = ri.RtPointer
    
    ri.RiArchiveRecord.argtypes = [RtToken, c_char_p]
    ri.RiAreaLightSource.argtypes = [RtToken]
    ri.RiAreaLightSource.restype = RtLightHandle 
    ri.RiAtmosphere.argtypes = [RtToken]
    ri.RiAttribute.argtypes = [RtToken]
    ri.RiAttributeBegin.argtypes = []
    ri.RiAttributeEnd.argtypes = []
    ri.RiBasis.argtypes = [RtBasis, RtInt, RtBasis, RtInt]
    ri.RiBegin.argtypes = [RtToken]
    ri.RiBlobby.argtypes = [RtInt, RtInt, POINTER(RtInt), RtInt, POINTER(RtFloat), RtInt, POINTER(RtString)]
    ri.RiBound.argtypes = [RtBound]
    ri.RiClipping.argtypes = [RtFloat, RtFloat]
    ri.RiClippingPlane.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiColor.argtypes = [RtColor]
    ri.RiColorSamples.argtypes = [RtInt, POINTER(RtFloat), POINTER(RtFloat)]
    ri.RiConcatTransform.argtypes = [RtMatrix]
    ri.RiCone.argtypes = [RtFloat, RtFloat, RtFloat]
    ri.RiContext.argtypes = [RtContextHandle]
    ri.RiCoordinateSystem.argtypes = [RtToken]
    ri.RiCoordSysTransform.argtypes = [RtToken]
    ri.RiCropWindow.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiCurves.argtypes = [RtToken, RtInt]
    ri.RiCylinder.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiDeclare.argtypes = [c_char_p, c_char_p]
    ri.RiDeclare.restype = RtToken
    ri.RiDepthOfField.argtypes = [RtFloat, RtFloat, RtFloat]
    ri.RiDetail.argtypes = [RtBound]
    ri.RiDetailRange.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiDisk.argtypes = [RtFloat, RtFloat, RtFloat]
    ri.RiDisplacement.argtypes = [RtToken]
    ri.RiDisplay.argtypes = [RtToken, RtToken, RtToken]
    ri.RiEnd.argtypes = []
    # See RiPixelFilter for an explanation why RiErrorHandler is commented out
#    ri.RiErrorHandler.argtypes = [RtErrorHandler]
    ri.RiExposure.argtypes = [RtFloat, RtFloat]
    ri.RiExterior.argtypes = [RtToken]
    ri.RiFormat.argtypes = [RtInt, RtInt, RtFloat]
    ri.RiFrameAspectRatio.argtypes = [RtFloat]
    ri.RiFrameBegin.argtypes = [RtInt]
    ri.RiFrameEnd.argtypes = []
    ri.RiGeneralPolygon.argtypes = [RtInt, POINTER(RtInt)]
    ri.RiGeometricApproximation.argtypes = [RtToken, RtFloat]
    ri.RiGeometry.argtypes = [RtToken]
    ri.RiGetContext.argtypes = []
    ri.RiGetContext.restype = RtContextHandle
    ri.RiHider.argtypes = [RtToken]
    ri.RiHyperboloid.argtypes = [RtPoint, RtPoint, RtFloat]
    ri.RiIdentity.argtypes = []
    ri.RiIlluminate.argtypes = [RtLightHandle, RtBoolean]
    ri.RiImager.argtypes = [RtToken]
    ri.RiInterior.argtypes = [RtToken]
    ri.RiLightSource.argtypes = [RtToken]
    ri.RiLightSource.restype = RtLightHandle
    # In the following texture calls the declaration of the filter function is removed
    # (see RiPixelFilter for more infos)
#    ri.RiMakeCubeFaceEnvironment.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, RtFloat, RtFilterFunc, RtFloat, RtFloat]
    ri.RiMakeCubeFaceEnvironment.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, c_char_p, RtFloat]
#    ri.RiMakeLatLongEnvironment.argtypes = [c_char_p, c_char_p, RtFilterFunc, RtFloat, RtFloat]
    ri.RiMakeLatLongEnvironment.argtypes = [c_char_p, c_char_p]
    ri.RiMakeShadow.argtypes = [c_char_p, c_char_p]
#    ri.RiMakeTexture.argtypes = [c_char_p, c_char_p, RtToken, RtToken, RtFilterFunc, RtFloat, RtFloat]
    ri.RiMakeTexture.argtypes = [c_char_p, c_char_p, RtToken, RtToken]
    ri.RiMatte.argtypes = [RtBoolean]
    ri.RiMotionBegin.argtypes = [RtInt]
    ri.RiMotionEnd.argtypes = []
    ri.RiNuPatch.argtypes = [RtInt, RtInt, POINTER(RtFloat), RtFloat, RtFloat, RtInt, RtInt, POINTER(RtFloat), RtFloat, RtFloat]
    ri.RiObjectBegin.argtypes = []
    ri.RiObjectBegin.restype = RtObjectHandle
    ri.RiObjectEnd.argtypes = []
    ri.RiObjectInstance.argtypes = [RtObjectHandle]
    ri.RiOpacity.argtypes = [RtColor]
    ri.RiOption.argtypes = [RtToken]
    ri.RiOrientation.argtypes = [RtToken]
    ri.RiParaboloid.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiPatch.argtypes = [RtToken]
    ri.RiPatchMesh.argtypes = [RtToken, RtInt, RtToken, RtInt, RtToken]
    ri.RiPerspective.argtypes = [RtFloat]
    # The following line is commented out because declaring the first argument
    # as RtFilterFunc will prevent the standard filters from being passed in directly
    # so that the renderer recognizes them as standard filters (which is necessary
    # during RIB generation so that the renderer can generate the appropriate
    # filter name).
#    ri.RiPixelFilter.argtypes = [RtFilterFunc, RtFloat, RtFloat]
    ri.RiPixelSamples.argtypes = [RtFloat, RtFloat]
    ri.RiPixelVariance.argtypes = [RtFloat]
    ri.RiPoints.argtypes = [RtInt]
    ri.RiPointsGeneralPolygons.argtypes = [RtInt, POINTER(RtInt), POINTER(RtInt), POINTER(RtInt)]
    ri.RiPointsPolygons.argtypes = [RtInt, POINTER(RtInt), POINTER(RtInt)]
    ri.RiPolygon.argtypes = [RtInt]
#    ri.RiProcedural.argtypes = [RtPointer, RtBound, RtProcSubdivFunc, RtProcFreeFunc]
    ri.RiProcedural.argtypes = [RtPointer, RtBound]
    ri.RiProjection.argtypes = [RtToken]
    ri.RiQuantize.argtypes = [RtToken, RtInt, RtInt, RtInt, RtFloat]
    # When the second argument is set to RtArchiveCallback then it is not possible to pass None
#    ri.RiReadArchive.argtypes = [RtToken, RtArchiveCallback]
    ri.RiReadArchive.argtypes = [RtToken]
    ri.RiRelativeDetail.argtypes = [RtFloat]
    ri.RiReverseOrientation.argtypes = []
    ri.RiRotate.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiScale.argtypes = [RtFloat, RtFloat, RtFloat]
    ri.RiScreenWindow.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiShadingInterpolation.argtypes = [RtToken]
    ri.RiShadingRate.argtypes = [RtFloat]
    ri.RiShutter.argtypes = [RtFloat, RtFloat]
    ri.RiSides.argtypes = [RtInt]
    ri.RiSkew.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiSolidBegin.argtypes = [RtToken]
    ri.RiSolidEnd.argtypes = []
    ri.RiSphere.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiSubdivisionMesh.argtypes = [RtToken, RtInt, POINTER(RtInt), POINTER(RtInt), RtInt, POINTER(RtToken), POINTER(RtInt), POINTER(RtInt), POINTER(RtFloat)]
    ri.RiSurface.argtypes = [RtToken]
    ri.RiTextureCoordinates.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat, RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiTorus.argtypes = [RtFloat, RtFloat, RtFloat, RtFloat, RtFloat]
    ri.RiTransform.argtypes = [RtMatrix]
    ri.RiTransformBegin.argtypes = []
    ri.RiTransformEnd.argtypes = []
    ri.RiTransformPoints.argtypes = [RtToken, RtToken, RtInt, POINTER(RtPoint)]
    ri.RiTranslate.argtypes = [RtFloat, RtFloat, RtFloat]
    ri.RiTrimCurve.argtypes = [RtInt, POINTER(RtInt), POINTER(RtInt), POINTER(RtFloat), POINTER(RtFloat), POINTER(RtFloat), POINTER(RtInt), POINTER(RtFloat), POINTER(RtFloat), POINTER(RtFloat)]
    ri.RiWorldBegin.argtypes = []
    ri.RiWorldEnd.argtypes = []
    
    ri.RiBoxFilter.argtypes = [c_float, c_float, c_float, c_float]
    ri.RiBoxFilter.restype = c_float
    ri.RiTriangleFilter.argtypes = [c_float, c_float, c_float, c_float]
    ri.RiTriangleFilter.restype = c_float
    ri.RiCatmullRomFilter.argtypes = [c_float, c_float, c_float, c_float]
    ri.RiCatmullRomFilter.restype = c_float
    ri.RiGaussianFilter.argtypes = [c_float, c_float, c_float, c_float]
    ri.RiGaussianFilter.restype = c_float
    ri.RiSincFilter.argtypes = [c_float, c_float, c_float, c_float]
    ri.RiSincFilter.restype = c_float
    
    # Optional calls from the 3.4 API:
    
    if hasattr(ri, "RiArchiveBegin"):
        ri.RiArchiveBegin.argtypes = [RtToken]
        ri.RiArchiveBegin.restype = c_void_p
    else:
        ri.RiArchiveBegin = _undefinedFunctionSubstitute

    if hasattr(ri, "RiArchiveEnd"):
        ri.RiArchiveEnd.argtypes = []
    else:
        ri.RiArchiveEnd = _undefinedFunctionSubstitute

    if hasattr(ri, "RiCamera"):
        ri.RiCamera.argtypes = [RtToken]
    else:
        ri.RiCamera = _undefinedFunctionSubstitute

    if hasattr(ri, "RiDisplayChannel"):
        ri.RiDisplayChannel.argtypes = [RtToken]
    else:
        ri.RiDisplayChannel = _undefinedFunctionSubstitute

    if hasattr(ri, "RiElse"):
        ri.RiElse.argtypes = []
    else:
        ri.RiElse = _undefinedFunctionSubstitute

    if hasattr(ri, "RiElseIf"):
        ri.RiElseIf.argtypes = [RtToken]
    else:
        ri.RiElseIf = _undefinedFunctionSubstitute
        
    if hasattr(ri, "RiIfBegin"):
        ri.RiIfBegin.argtypes = [RtToken]
    else:
        ri.RiIfBegin = _undefinedFunctionSubstitute

    if hasattr(ri, "RiIfEnd"):
        ri.RiIfEnd.argtypes = []
    else:
        ri.RiIfEnd = _undefinedFunctionSubstitute

    if hasattr(ri, "RiMakeBrickMap"):
        ri.RiMakeBrickMap.argtypes = [RtInt, POINTER(c_char_p), c_char_p]
    else:
        ri.RiMakeBrickMap = _undefinedFunctionSubstitute

    if hasattr(ri, "RiResource"):
        ri.RiResource.argtypes = [RtToken, RtToken]
    else:
        ri.RiResource = _undefinedFunctionSubstitute

    if hasattr(ri, "RiResourceBegin"):
        ri.RiResourceBegin.argtypes = []
    else:
        ri.RiResourceBegin = _undefinedFunctionSubstitute

    if hasattr(ri, "RiResourceEnd"):
        ri.RiResourceEnd.argtypes = []
    else:
        ri.RiResourceEnd = _undefinedFunctionSubstitute

    if hasattr(ri, "RiScopedCoordinateSystem"):
        ri.RiScopedCoordinateSystem.argtypes = [RtToken]
    else:
        ri.RiScopedCoordinateSystem = _undefinedFunctionSubstitute

    if hasattr(ri, "RiShader"):
        ri.RiShader.argtypes = [RtToken, RtToken]
    else:
        ri.RiShader = _undefinedFunctionSubstitute

    if hasattr(ri, "RiSystem"):
        ri.RiSystem.argtypes = [c_char_p]
    else:
        ri.RiSystem = _undefinedFunctionSubstitute
    
def _undefinedFunctionSubstitute(*args, **kwargs):
    raise NotImplementedError("This function call is not implemented by this implementation of the RenderMan interface")
    
def _getLastError(ri):
    """Getter function for the RiLastError variable.
    """
    try:
        return c_int.in_dll(ri, "RiLastError").value
    except ValueError:
        return None    

def _setLastError(ri, value):
    """Setter function for the RiLastError variable.
    """
    try:
        c_int.in_dll(ri, "RiLastError").value = value
    except ValueError:
        pass
    
   



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: timer.py,v 1.1.1.1 2004/12/12 14:31:27 mbaas Exp $

## \file timer.py
## Contains the Timer class.

import time
from .component import Component
from .eventmanager import eventManager
from . import slots, events

# Timer
class Timer(Component):
    """%Timer class that manages the virtual time.

    Attributes:

    - time (Slot) - Current virtual time
    - timestep - Delta time step
    - duration - Total duration of the animation/simulation
    
    - frame - Current frame number (as float)
    - fps - Frame rate (as float)
    - framecount - Total number of frames (as float)
    
    - clock - Contains the real time (this is a real stop watch)

    The actual slot object of each slot attribute can be accessed by appending
    "_slot" the attribute name.
    """

    def __init__(self, name="Timer", auto_insert=True):
        Component.__init__(self, name=name, auto_insert=auto_insert)

        # The time() value when the clock was started
        self._clock_start  = 0.0
        # This offset is added to the returned clock value. It stores the
        # clock value at the time the clock was stopped
        self._clock_offset = 0.0
        # This flag indicates if the clock is running or not
        self._clock_running = False

        self._duration = 1.0
        self._framecount = 25
        self._fps = 25
        self._timestep = 0.04
        
        # Virtual time...
        self.time_slot = slots.DoubleSlot()
        self.addSlot("time", self.time_slot)

        eventManager().connect(events.RESET, self)
        

    def startClock(self):
        self._clock_start = time.time()
        self._clock_running = True        

    def stopClock(self):
        self._clock_offset = self.clock
        self._clock_running = False

    def step(self):
        t = self.time_slot.getValue() + self._timestep
        self.time_slot.setValue(t)
        eventManager().event(events.STEP_FRAME)

    ## protected:

    def onReset(self):
        """Reset callback."""
        self.time = 0.0
        
    # "time" property...
    
    def _getTime(self):
        """Return the current virtual time.

        This method is used for retrieving the \a time property.

        \return Virtual time (\c float)
        """
        return self.time_slot.getValue()

    def _setTime(self, t):
        """Set the virtual time.

        This method is used for setting the \a time property.

        \param t (\c float) Time value
        """
        self.time_slot.setValue(t)

    time = property(_getTime, _setTime, None, "Virtual time")

    # "frame" property...
    
    def _getFrame(self):
        """Return the current frame number.

        This method is used for retrieving the \a frame property.

        \return Frame number (\c float)
        """
        return self.time_slot.getValue()/self._timestep

    def _setFrame(self, f):
        """Set the frame number.

        This method is used for setting the \a frame property.

        \param f (\c float) Frame number
        """
        self.time_slot.setValue(f*self._timestep)

    frame = property(_getFrame, _setFrame, None, "Frame number")

    # "clock" property...

    def _getClock(self):
        """Return the current clock value (real time).

        This method is used for retrieving the \a clock property.

        \return Real time (\c float)
        """
        if self._clock_running:
            return self._clock_offset + time.time()-self._clock_start
        else:
            return self._clock_offset

    def _setClock(self, t):
        """Set the clock value.

        This method is used for setting the \a clock property.

        \param t (\c float) Time value
        """
        self._clock_offset = float(t)
        self._clock_start = time.time()

    clock = property(_getClock, _setClock, None, "Real time")

    # "timestep" property...
    
    def _getTimeStep(self):
        """Return the current time step.

        This method is used for retrieving the \a timestep property.

        \return Time step (\c float)
        """
        return self._timestep

    def _setTimeStep(self, dt):
        """Set the time step.

        This method is used for setting the \a timestep property.

        \param t (\c float) Time value
        """
        self._timestep = float(dt)

    timestep = property(_getTimeStep, _setTimeStep, None, "Time step")

    # "fps" property...
    
    def _getFPS(self):
        """Return the current frame rate.

        This method is used for retrieving the \a fps property.

        \return Frame rate (\c float)
        """
        return 1.0/self._timestep

    def _setFPS(self, fps):
        """Set the frame rate.

        This method is used for setting the \a fps property.

        \param fps (\c float) New frame rate
        """
        self._timestep = 1.0/float(fps)

    fps = property(_getFPS, _setFPS, None, "Frames per second")

    # "duration" property...
    
    def _getDuration(self):
        """Return the duration of the animation/simulation.

        This method is used for retrieving the \a duration property.

        \return Duration (\c float)
        """
        return self._duration

    def _setDuration(self, d):
        """Set the duration.

        This method is used for setting the \a duration property.

        \param d (\c float) New duration
        """
        self._duration = float(d)

    duration = property(_getDuration, _setDuration, None, "Total duration")

    # "framecount" property...
    
    def _getFrameCount(self):
        """Return the total number of frames.

        This method is used for retrieving the \a framecount property.

        \return Total Number of frames (\c float)
        """
        return self._duration/self._timestep

    def _setFrameCount(self, n):
        """Set the total number of frames.

        This method is used for setting the \a framecount property.

        \param n (\c float) Total number of frames
        """
        self._duration = float(n)*self._timestep

    framecount = property(_getFrameCount, _setFrameCount, None, "Total number of frames")



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: vec3.py,v 1.1 2005/08/15 15:39:48 mbaas Exp $

import sys
import math

# Comparison threshold
_epsilon = 1E-12

# vec3
class vec3:
    """Three-dimensional vector.

    This class can be used to represent points, vectors, normals
    or even colors. The usual vector operations are available.
    """

    def __init__(self, *args):
        """Constructor.

        There are several possibilities how to initialize a vector:

        v = vec3()       -> v = <0,0,0>
        v = vec3(a)      -> v = <a,a,a>
        v = vec3(x,y)    -> v = <x,y,0>
        v = vec3(x,y,z)  -> v = <x,y,z>

        Note that specifying just one value sets all three components to
        that value (except when that single value is a another vec3, then
        that vector is copied).

        Additionally you can wrap those values in a list or a tuple or
        specify them as a string:

        v = vec3([1,2,3]) -> v = <1,2,3>
        v = vec3("4,5")   -> v = <4,5,0>        
        """
        
        if len(args)==0:
            self.x, self.y, self.z = (0.0, 0.0, 0.0)

        elif len(args)==1:
            T = type(args[0])
            # scalar
            if T in [int, float]:
                f = float(args[0])
                self.x, self.y, self.z = (f, f, f)
            # vec3  
            elif isinstance(args[0], vec3):
                self.x, self.y, self.z = args[0]
            # Tuple/List
            elif T is tuple or T is list:
                if len(args[0])==0:
                    self.x = self.y = self.z = 0.0
                elif len(args[0])==1:
                    self.x = self.y = self.z = float(args[0][0])
                elif len(args[0])==2:
                    self.x = float(args[0][0])
                    self.y = float(args[0][1])
                    self.z = 0.0
                elif len(args[0])==3:
                    x,y,z = args[0]
                    self.x = float(x)
                    self.y = float(y)
                    self.z = float(z)
                else:
                    raise TypeError("vec3() takes at most 3 arguments")
            # String
            elif T is str:
                s=args[0].replace(","," ").replace("  "," ").strip().split(" ")
                if s==[""]:
                    s=[]
                f=list([float(x) for x in s])
                dummy = vec3(f)
                self.x, self.y, self.z = dummy
            # error
            else:
                raise TypeError("vec3() arg can't be converted to vec3")

        elif len(args)==2:
            self.x, self.y, self.z = (float(args[0]), float(args[1]), 0.0)
            
        elif len(args)==3:
            x,y,z = args
            self.x = float(x)
            self.y = float(y)
            self.z = float(z)

        else:
            raise TypeError("vec3() takes at most 3 arguments")


    def __repr__(self):
        return 'vec3(%r, %r, %r)'%(self.x, self.y, self.z)

    def __str__(self):
        return '(%1.4f, %1.4f, %1.4f)'%(self.x, self.y, self.z)

    # Make the object unhashable (as it is mutable)
    __hash__ = None

    def __eq__(self, other):
        """== operator

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> b=vec3(-0.3, 0.75, 0.5)
        >>> c=vec3(-0.3, 0.75, 0.5)
        >>> print a==b
        0
        >>> print b==c
        1
        >>> print a==None
        0
        """
        global _epsilon
        if isinstance(other, vec3):
            return (abs(self.x-other.x)<=_epsilon and
                    abs(self.y-other.y)<=_epsilon and
                    abs(self.z-other.z)<=_epsilon)
        else:
            return False

    def __ne__(self, other):
        """!= operator

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> b=vec3(-0.3, 0.75, 0.5)
        >>> c=vec3(-0.3, 0.75, 0.5)
        >>> print a!=b
        1
        >>> print b!=c
        0
        >>> print a!=None
        1
        """
        return not (self==other)

    def __lt__(self, other):
        """< operator."""
        if isinstance(other, vec3):
            return ((self.x<other.x) and
                    (self.y<other.y) and
                    (self.z<other.z))
        else:
            return False

    def __le__(self, other):
        """<= operator."""
        if isinstance(other, vec3):
            return ((self.x<=other.x) and
                    (self.y<=other.y) and
                    (self.z<=other.z))
        else:
            return False
        
    def __gt__(self, other):
        """> operator."""
        if isinstance(other, vec3):
            return ((self.x>other.x) and
                    (self.y>other.y) and
                    (self.z>other.z))
        else:
            return False
        
    def __ge__(self, other):
        """>= operator."""
        if isinstance(other, vec3):
            return ((self.x>=other.x) and
                    (self.y>=other.y) and
                    (self.z>=other.z))
        else:
            return False


    def __add__(self, other):
        """Vector addition.

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> b=vec3(-0.3, 0.75, 0.5)
        >>> print a+b
        (0.7000, 1.2500, -1.3000)
        """
        if isinstance(other, vec3):
            return vec3(self.x+other.x, self.y+other.y, self.z+other.z)
        else:
            raise TypeError("unsupported operand type for +")

    def __sub__(self, other):
        """Vector subtraction.

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> b=vec3(-0.3, 0.75, 0.5)
        >>> print a-b
        (1.3000, -0.2500, -2.3000)
        """
        if isinstance(other, vec3):
            return vec3(self.x-other.x, self.y-other.y, self.z-other.z)
        else:
            raise TypeError("unsupported operand type for -")

    def __mul__(self, other):
        """Multiplication with a scalar or dot product.

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> b=vec3(-0.3, 0.75, 0.5)
        >>> print a*2.0
        (2.0000, 1.0000, -3.6000)
        >>> print 2.0*a
        (2.0000, 1.0000, -3.6000)
        >>> print a*b
        -0.825
        """
        
        T = type(other)
        # vec3*scalar
        if T in [int, float]:
            return vec3(self.x*other, self.y*other, self.z*other)
        # vec3*vec3
        if isinstance(other, vec3):
            return self.x*other.x + self.y*other.y + self.z*other.z
        # unsupported
        else:
            # Try to delegate the operation to the other operand
            if getattr(other,"__rmul__",None)!=None:
                return other.__rmul__(self)
            else:
                raise TypeError("unsupported operand type for *")

    __rmul__ = __mul__

    def __truediv__(self, other):
        """Division by scalar

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> print a/2.0
        (0.5000, 0.2500, -0.9000)
        """
        T = type(other)
        # vec3/scalar
        if T in [int, float]:
            return vec3(self.x/other, self.y/other, self.z/other)
        # unsupported
        else:
            raise TypeError("unsupported operand type for /")

    # For Python <3:
    __div__ = __truediv__

    def __mod__(self, other):
        """Modulo (component wise)

        >>> a=vec3(3.0, 2.5, -1.8)
        >>> print a%2.0
        (1.0000, 0.5000, 0.2000)
        """
        T = type(other)
        # vec3%scalar
        if T in [int, float]:
            return vec3(self.x%other, self.y%other, self.z%other)
        # vec3%vec3
        if isinstance(other, vec3):
            return vec3(self.x%other.x, self.y%other.y, self.z%other.z)
        # unsupported
        else:
            raise TypeError("unsupported operand type for %")

    def __iadd__(self, other):
        """Inline vector addition.

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> b=vec3(-0.3, 0.75, 0.5)
        >>> a+=b
        >>> print a
        (0.7000, 1.2500, -1.3000)
        """
        if isinstance(other, vec3):
            self.x+=other.x
            self.y+=other.y
            self.z+=other.z
            return self
        else:
            raise TypeError("unsupported operand type for +=")

    def __isub__(self, other):
        """Inline vector subtraction.

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> b=vec3(-0.3, 0.75, 0.5)
        >>> a-=b
        >>> print a
        (1.3000, -0.2500, -2.3000)
        """
        if isinstance(other, vec3):
            self.x-=other.x
            self.y-=other.y
            self.z-=other.z
            return self
        else:
            raise TypeError("unsupported operand type for -=")

    def __imul__(self, other):
        """Inline multiplication (only with scalar)

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> a*=2.0
        >>> print a
        (2.0000, 1.0000, -3.6000)
        """
        x = float(self.x*other)
        y = float(self.y*other)
        z = float(self.z*other)
        self.x,self.y,self.z = x,y,z
        return self

    def __itruediv__(self, other):
        """Inline division with scalar

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> a/=2.0
        >>> print a
        (0.5000, 0.2500, -0.9000)
        """
        x = float(self.x/other)
        y = float(self.y/other)
        z = float(self.z/other)
        self.x,self.y,self.z = x,y,z
        return self

    # For Python <3:
    __idiv__ = __itruediv__

    def __imod__(self, other):
        """Inline modulo

        >>> a=vec3(3.0, 2.5, -1.8)
        >>> a%=2.0
        >>> print a
        (1.0000, 0.5000, 0.2000)
        """
        T = type(other)
        # vec3%=scalar
        if T in [int, float]:
            self.x%=other
            self.y%=other
            self.z%=other
            return self
        # vec3%vec3
        if isinstance(other, vec3):
            self.x%=other.x
            self.y%=other.y
            self.z%=other.z
            return self
        else:
            raise TypeError("unsupported operand type for %=")

    def __neg__(self):
        """Negation

        >>> a=vec3(3.0, 2.5, -1.8)
        >>> print -a
        (-3.0000, -2.5000, 1.8000)
        """
        return vec3(-self.x, -self.y, -self.z)

    def __pos__(self):
        """
        >>> a=vec3(3.0, 2.5, -1.8)
        >>> print +a
        (3.0000, 2.5000, -1.8000)
        """
        return vec3(+self.x, +self.y, +self.z)

    def __abs__(self):
        """Return the length of the vector.

        abs(v) is equivalent to v.length().

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> print abs(a)
        2.11896201004
        """
        return math.sqrt(self*self)


    def __len__(self):
        """Length of the sequence (always 3)"""
        return 3

    def __getitem__(self, key):
        """Return a component by index (0-based)

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> print a[0]
        1.0
        >>> print a[1]
        0.5
        >>> print a[2]
        -1.8
        """
        
        T=type(key)
        if T is not int:
            raise TypeError("index must be integer")

        if   key==0: return self.x
        elif key==1: return self.y
        elif key==2: return self.z
        else:
            raise IndexError("index out of range")

    def __setitem__(self, key, value):
        """Set a component by index (0-based)

        >>> a=vec3()
        >>> a[0]=1.5; a[1]=0.7; a[2]=-0.3
        >>> print a
        (1.5000, 0.7000, -0.3000)
        """
        
        T=type(key)
        if T is not int:
            raise TypeError("index must be integer")

        if   key==0: self.x = value
        elif key==1: self.y = value
        elif key==2: self.z = value
        else:
            raise IndexError("index out of range")

    def cross(self, other):
        """Cross product.

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> b=vec3(-0.3, 0.75, 0.5)
        >>> c=a.cross(b)
        >>> print c
        (1.6000, 0.0400, 0.9000)
        """
        
        if isinstance(other, vec3):
            return vec3(self.y*other.z-self.z*other.y,
                        self.z*other.x-self.x*other.z,
                        self.x*other.y-self.y*other.x)
        else:
            raise TypeError("unsupported operand type for cross()")
        

    def length(self):
        """Return the length of the vector.

        v.length() is equivalent to abs(v).

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> print a.length()
        2.11896201004
        """

        return math.sqrt(self*self)

    def normalize(self):
        """Return normalized vector.

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> print a.normalize()
        (0.4719, 0.2360, -0.8495)
        """

        nlen = 1.0/math.sqrt(self*self)
        return vec3(self.x*nlen, self.y*nlen, self.z*nlen)

    def angle(self, other):
        """Return angle (in radians) between self and other.

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> b=vec3(-0.3, 0.75, 0.5)
        >>> print a.angle(b)
        1.99306755584
        """

        other = vec3(other)
        return math.acos((self*other) / (abs(self)*abs(other)))

    def reflect(self, N):
        """Return the reflection vector.

        N is the surface normal which has to be of unit length.

        >>> a=vec3(1.0, 0.5, -1.8)
        >>> print a.reflect(vec3(1,0,1))
        (2.6000, 0.5000, -0.2000)
        """

        N = vec3(N)
        return self - 2.0*(self*N)*N

    def refract(self, N, eta):
        """Return the transmitted vector.

        N is the surface normal which has to be of unit length.
        eta is the relative index of refraction. If the returned
        vector is zero then there is no transmitted light because
        of total internal reflection.
        
        >>> a=vec3(1.0, -1.5, 0.8)
        >>> print a.refract(vec3(0,1,0), 1.33)
        (1.3300, -1.7920, 1.0640)
        """

        N = vec3(N)
        dot = self*N
        k   = 1.0 - eta*eta*(1.0 - dot*dot)
        if k<0:
            return vec3(0.0,0.0,0.0)
        else:
            return eta*self - (eta*dot + math.sqrt(k))*N

    def ortho(self):
        """Returns an orthogonal vector.

        Returns a vector that is orthogonal to self (where
        self*self.ortho()==0).

        >>> a=vec3(1.0, -1.5, 0.8)
        >>> print round(a*a.ortho(),8)
        0.0
        """

        x=abs(self.x)
        y=abs(self.y)
        z=abs(self.z)
        # Is x the smallest element? 
        if x<y and x<z:
            return vec3(0.0, -self.z, self.y)
        # Is y smallest element?
        elif y<z:
            return vec3(-self.z, 0.0, self.x)
        # z is smallest
        else:
            return vec3(-self.y, self.x, 0.0)

    def min(self):
        """Return the minimum value of the components.
        """
        return min(self.x, self.y, self.z)

    def max(self):
        """Return the maximum value of the components.
        """
        return max(self.x, self.y, self.z)

    def minIndex(self):
        """Return the index of the component with the minimum value.
        """
        if self.x<=self.y and self.x<=self.z:
            return 0
        elif self.y<=self.z:
            return 1
        else:
            return 2

    def maxIndex(self):
        """Return the index of the component with the maximum value.
        """
        if self.x>=self.y and self.x>=self.z:
            return 0
        elif self.y>=self.z:
            return 1
        else:
            return 2
        
    def minAbs(self):
        """Return the minimum absolute value of the components.
        """
        return min(abs(self.x), abs(self.y), abs(self.z))

    def maxAbs(self):
        """Return the maximum absolute value of the components.
        """
        return max(abs(self.x), abs(self.y), abs(self.z))

    def minAbsIndex(self):
        """Return the index of the component with the minimum absolute value.
        """
        ax = abs(self.x)
        ay = abs(self.y)
        az = abs(self.z)
        
        if ax<=ay and ax<=az:
            return 0
        elif ay<=az:
            return 1
        else:
            return 2

    def maxAbsIndex(self):
        """Return the index of the component with the maximum absolute value.
        """
        ax = abs(self.x)
        ay = abs(self.y)
        az = abs(self.z)

        if ax>=ay and ax>=az:
            return 0
        elif ay>=az:
            return 1
        else:
            return 2


######################################################################
def _test():
    import doctest, vec3
    failed, total = doctest.testmod(vec3)
    print(("%d/%d failed" % (failed, total)))

if __name__=="__main__":

    _test()

#    a = vec3(1,2,3.03)
#    b = vec3("-2,0.5,1E10")

#    print a.angle(b)
